# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Reinforcement learning loss functions.

All the loss functions implemented here compute the loss for the policy (actor).
The critic loss functions are typically regression loss are omitted for their
simplicity.

For the batch QPG, RM and RPG loss, please refer to the paper:
https://papers.nips.cc/paper/7602-actor-critic-policy-optimization-in-partially-observable-multiagent-environments.pdf

The BatchA2C loss uses code from the `TRFL` library:
https://github.com/deepmind/trfl/blob/master/trfl/discrete_policy_gradient_ops.py
"""

import torch
import torch.nn.functional as F


def _assert_rank_and_shape_compatibility(tensors, rank):
  if not tensors:
    raise ValueError("List of tensors cannot be empty")

  tmp_shape = tensors[0].shape
  for tensor in tensors:
    if tensor.ndim != rank:
      raise ValueError("Shape %s must have rank %d" % (tensor.ndim, rank))
    if tensor.shape != tmp_shape:
      raise ValueError("Shapes %s and %s are not compatible" %
                       (tensor.shape, tmp_shape))


def thresholded(logits, regrets, threshold=2.0):
  """Zeros out `regrets` where `logits` are too negative or too large."""
  can_decrease = logits.gt(-threshold).float()
  can_increase = logits.lt(threshold).float()
  regrets_negative = regrets.clamp(max=0.0)
  regrets_positive = regrets.clamp(min=0.0)
  return can_decrease * regrets_negative + can_increase * regrets_positive


def compute_baseline(policy, action_values):
  # V = pi * Q, backprop through pi but not Q.
  return torch.sum(torch.mul(policy, action_values.detach()), dim=1)


def compute_regrets(policy_logits, action_values):
  """Compute regrets using pi and Q."""
  # Compute regret.
  policy = F.softmax(policy_logits, dim=1)
  # Avoid computing gradients for action_values.
  action_values = action_values.detach()

  baseline = compute_baseline(policy, action_values)

  regrets = torch.sum(
      F.relu(action_values - torch.unsqueeze(baseline, 1)), dim=1)

  return regrets


def compute_advantages(policy_logits,
                       action_values,
                       use_relu=False,
                       threshold_fn=None):
  """Compute advantages using pi and Q."""
  # Compute advantage.
  policy = F.softmax(policy_logits, dim=1)
  # Avoid computing gradients for action_values.
  action_values = action_values.detach()

  baseline = compute_baseline(policy, action_values)

  advantages = action_values - torch.unsqueeze(baseline, 1)
  if use_relu:
    advantages = F.relu(advantages)

  if threshold_fn:
    # Compute thresholded advanteges weighted by policy logits for NeuRD.
    policy_logits = policy_logits - policy_logits.mean(-1, keepdim=True)
    advantages = threshold_fn(policy_logits, advantages)
    policy_advantages = -torch.mul(policy_logits, advantages.detach())
  else:
    # Compute advantage weighted by policy.
    policy_advantages = -torch.mul(policy, advantages.detach())
  return torch.sum(policy_advantages, dim=1)


def compute_a2c_loss(policy_logits, actions, advantages):
  cross_entropy = F.cross_entropy(policy_logits, actions, reduction="none")
  advantages = advantages.detach()
  if advantages.ndim != cross_entropy.ndim:
    raise ValueError("Shapes %s and %s are not compatible" %
                     (advantages.ndim, cross_entropy.ndim))
  return torch.mul(cross_entropy, advantages)


def compute_entropy(policy_logits):
  return torch.sum(
      -F.softmax(policy_logits, dim=1) * F.log_softmax(policy_logits, dim=1),
      dim=-1)


class BatchQPGLoss(object):
  """Defines the batch QPG loss op."""

  def __init__(self, entropy_cost=None, name="batch_qpg_loss"):
    self._entropy_cost = entropy_cost
    self._name = name

  def loss(self, policy_logits, action_values):
    """Constructs a PyTorch Crierion that computes the QPG loss for batches.

    Args:
      policy_logits: `B x A` tensor corresponding to policy logits.
      action_values: `B x A` tensor corresponding to Q-values.

    Returns:
      loss: A 0-D `float` tensor corresponding the loss.
    """
    _assert_rank_and_shape_compatibility([policy_logits, action_values], 2)
    advantages = compute_advantages(policy_logits, action_values)
    _assert_rank_and_shape_compatibility([advantages], 1)
    total_adv = torch.mean(advantages, dim=0)

    total_loss = total_adv
    if self._entropy_cost:
      policy_entropy = torch.mean(compute_entropy(policy_logits))
      entropy_loss = torch.mul(float(self._entropy_cost), policy_entropy)
      total_loss = torch.add(total_loss, entropy_loss)

    return total_loss


class BatchNeuRDLoss(object):
  """Defines the batch NeuRD loss op."""

  def __init__(self, entropy_cost=None, name="batch_neurd_loss"):
    self._entropy_cost = entropy_cost
    self._name = name

  def loss(self, policy_logits, action_values):
    """Constructs a PyTorch Crierion that computes the NeuRD loss for batches.

    Args:
      policy_logits: `B x A` tensor corresponding to policy logits.
      action_values: `B x A` tensor corresponding to Q-values.

    Returns:
      loss: A 0-D `float` tensor corresponding the loss.
    """
    _assert_rank_and_shape_compatibility([policy_logits, action_values], 2)
    advantages = compute_advantages(
        policy_logits, action_values, threshold_fn=thresholded)
    _assert_rank_and_shape_compatibility([advantages], 1)
    total_adv = torch.mean(advantages, axis=0)

    total_loss = total_adv
    if self._entropy_cost:
      policy_entropy = torch.mean(compute_entropy(policy_logits))
      entropy_loss = torch.mul(float(self._entropy_cost), policy_entropy)
      total_loss = torch.add(total_loss, entropy_loss)

    return total_loss


class BatchRMLoss(object):
  """Defines the batch RM loss op."""

  def __init__(self, entropy_cost=None, name="batch_rm_loss"):
    self._entropy_cost = entropy_cost
    self._name = name

  def loss(self, policy_logits, action_values):
    """Constructs a PyTorch Crierion that computes the RM loss for batches.

    Args:
      policy_logits: `B x A` tensor corresponding to policy logits.
      action_values: `B x A` tensor corresponding to Q-values.

    Returns:
      loss: A 0-D `float` tensor corresponding the loss.
    """
    _assert_rank_and_shape_compatibility([policy_logits, action_values], 2)
    advantages = compute_advantages(policy_logits, action_values, use_relu=True)
    _assert_rank_and_shape_compatibility([advantages], 1)
    total_adv = torch.mean(advantages, dim=0)

    total_loss = total_adv
    if self._entropy_cost:
      policy_entropy = torch.mean(compute_entropy(policy_logits))
      entropy_loss = torch.mul(float(self._entropy_cost), policy_entropy)
      total_loss = torch.add(total_loss, entropy_loss)

    return total_loss


class BatchRPGLoss(object):
  """Defines the batch RPG loss op."""

  def __init__(self, entropy_cost=None, name="batch_rpg_loss"):
    self._entropy_cost = entropy_cost
    self._name = name

  def loss(self, policy_logits, action_values):
    """Constructs a PyTorch Crierion that computes the RPG loss for batches.

    Args:
      policy_logits: `B x A` tensor corresponding to policy logits.
      action_values: `B x A` tensor corresponding to Q-values.

    Returns:
      loss: A 0-D `float` tensor corresponding the loss.
    """
    _assert_rank_and_shape_compatibility([policy_logits, action_values], 2)
    regrets = compute_regrets(policy_logits, action_values)
    _assert_rank_and_shape_compatibility([regrets], 1)
    total_regret = torch.mean(regrets, dim=0)

    total_loss = total_regret
    if self._entropy_cost:
      policy_entropy = torch.mean(compute_entropy(policy_logits))
      entropy_loss = torch.mul(float(self._entropy_cost), policy_entropy)
      total_loss = torch.add(total_loss, entropy_loss)

    return total_loss


class BatchA2CLoss(object):
  """Defines the batch A2C loss op."""

  def __init__(self, entropy_cost=None, name="batch_a2c_loss"):
    self._entropy_cost = entropy_cost
    self._name = name

  def loss(self, policy_logits, baseline, actions, returns):
    """Constructs a PyTorch Crierion that computes the A2C loss for batches.

    Args:
      policy_logits: `B x A` tensor corresponding to policy logits.
      baseline: `B` tensor corresponding to baseline (V-values).
      actions: `B` tensor corresponding to actions taken.
      returns: `B` tensor corresponds to returns accumulated.

    Returns:
      loss: A 0-D `float` tensor corresponding the loss.
    """
    _assert_rank_and_shape_compatibility([policy_logits], 2)
    _assert_rank_and_shape_compatibility([baseline, actions, returns], 1)
    advantages = returns - baseline

    policy_loss = compute_a2c_loss(policy_logits, actions, advantages)
    total_loss = torch.mean(policy_loss, dim=0)
    if self._entropy_cost:
      policy_entropy = torch.mean(compute_entropy(policy_logits))
      entropy_loss = torch.mul(float(self._entropy_cost), policy_entropy)
      total_loss = torch.add(total_loss, entropy_loss)

    return total_loss



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Some basic stats classes."""

import math
from typing import List


class BasicStats(object):
  """A set of statistics about a single value series."""
  __slots__ = ("_num", "_min", "_max", "_sum", "_sum_sq")

  def __init__(self):
    self.reset()

  def reset(self):
    self._num = 0
    self._min = float("inf")
    self._max = float("-inf")
    self._sum = 0
    self._sum_sq = 0

  def add(self, val: float):
    self._num += 1
    if self._min > val:
      self._min = val
    if self._max < val:
      self._max = val
    self._sum += val
    self._sum_sq += val**2

  @property
  def num(self):
    return self._num

  @property
  def min(self):
    return 0 if self._num == 0 else self._min

  @property
  def max(self):
    return 0 if self._num == 0 else self._max

  @property
  def avg(self):
    return 0 if self._num == 0 else self._sum / self._num

  @property
  def std_dev(self):
    """Standard deviation."""
    if self._num == 0:
      return 0
    return math.sqrt(
        max(0, self._sum_sq / self._num - (self._sum / self._num)**2))

  def merge(self, other: "BasicStats"):
    # pylint: disable=protected-access
    self._num += other._num
    self._min = min(self._min, other._min)
    self._max = max(self._max, other._max)
    self._sum += other._sum
    self._sum_sq += other._sum_sq
    # pylint: enable=protected-access

  @property
  def as_dict(self):
    return {
        "num": self.num,
        "min": self.min,
        "max": self.max,
        "avg": self.avg,
        "std_dev": self.std_dev,
    }

  def __str__(self):
    if self.num == 0:
      return "num=0"
    return "sum: %.4f, avg: %.4f, dev: %.4f, min: %.4f, max: %.4f, num: %d" % (
        self.sum, self.avg, self.dev, self.min, self.max, self.num)


class HistogramNumbered:
  """Track a histogram of occurences for `count` buckets.

  You need to decide how to map your data into the buckets. Mainly useful for
  scalar values.
  """

  def __init__(self, num_buckets: int):
    self._counts = [0] * num_buckets

  def reset(self):
    self._counts = [0] * len(self._counts)

  def add(self, bucket_id: int):
    self._counts[bucket_id] += 1

  @property
  def data(self):
    return self._counts


class HistogramNamed:
  """Track a histogram of occurences for named buckets.

  Same as HistogramNumbered, but each bucket has a name associated with it.
  Mainly useful for categorical values.
  """

  def __init__(self, bucket_names: List[str]):
    self._names = bucket_names
    self.reset()

  def reset(self):
    self._counts = [0] * len(self._names)

  def add(self, bucket_id: int):
    self._counts[bucket_id] += 1

  @property
  def data(self):
    return {
        "counts": self._counts,
        "names": self._names,
    }



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Train a policy net on Hearts actions based given a dataset of trajectories.

Trajectories from the Hearts bot Xinxin can be generated using
open_spiel/bots/xinxin/xinxin_game_generator.cc.
"""

import os
import pickle
from typing import Any, Tuple

from absl import app
from absl import flags

import haiku as hk
import jax
from jax import numpy as jnp
import numpy as np
import optax

import pyspiel

OptState = Any
Params = Any

FLAGS = flags.FLAGS
GAME = pyspiel.load_game('hearts')
NUM_CARDS = 52
NUM_ACTIONS = NUM_CARDS
NUM_PLAYERS = 4
TOP_K_ACTIONS = 5  # How many alternative actions to display
DEFAULT_LAYER_SIZES = [1024, 1024, 1024, 1024]

flags.DEFINE_integer('iterations', 100000, 'Number of iterations')
flags.DEFINE_string('data_path', None, 'Location for data')
flags.DEFINE_integer('eval_every', 10000, 'How often to evaluate the policy')
flags.DEFINE_integer('num_examples', 3,
                     'How many examples to print per evaluation')
flags.DEFINE_integer('train_batch', 128, 'Batch size for training step')
flags.DEFINE_integer('eval_batch', 10000, 'Batch size when evaluating')
flags.DEFINE_float('step_size', 1e-4, 'Step size for training')
flags.DEFINE_list('hidden_layer_sizes', None,
                  'Number of hidden units and layers in the network')
flags.DEFINE_integer('rng_seed', 42, 'Seed for initial network weights')
flags.DEFINE_string('save_path', None, 'Location for saved networks')
flags.DEFINE_string('checkpoint_file', None,
                    'Provides weights and optimzer state to resume training')


def _trajectory(line: str):
  """Returns parsed action trajectory."""
  actions = [int(x) for x in line.split(' ')]
  return tuple(actions)


def make_dataset(file: str):
  """Creates dataset as a generator of single examples."""
  lines = [line for line in open(file)]
  while True:
    np.random.shuffle(lines)
    for line in lines:
      trajectory = _trajectory(line)
      # skip pass_dir and deal actions
      action_index = np.random.randint(NUM_CARDS + 1, len(trajectory))
      state = GAME.new_initial_state()
      for action in trajectory[:action_index]:
        state.apply_action(action)
      yield (state.information_state_tensor(), trajectory[action_index])


def batch(dataset, batch_size: int):
  """Creates a batched dataset from a one-at-a-time dataset."""
  observations = np.zeros([batch_size] + GAME.information_state_tensor_shape(),
                          np.float32)
  labels = np.zeros(batch_size, dtype=np.int32)
  while True:
    for batch_index in range(batch_size):
      observations[batch_index], labels[batch_index] = next(dataset)
    yield observations, labels


def one_hot(x, k):
  """Returns a one-hot encoding of `x` of size `k`."""
  return jnp.array(x[..., jnp.newaxis] == jnp.arange(k), dtype=np.float32)


def net_fn(x):
  """Haiku module for our network."""
  layers = []
  for layer_size in FLAGS.hidden_layer_sizes:
    layers.append(hk.Linear(int(layer_size)))
    layers.append(jax.nn.relu)
  layers.append(hk.Linear(NUM_ACTIONS))
  layers.append(jax.nn.log_softmax)
  net = hk.Sequential(layers)
  return net(x)


def main(argv):
  if len(argv) > 1:
    raise app.UsageError('Too many command-line arguments.')

  if FLAGS.hidden_layer_sizes is None:
    # Cannot pass default arguments as lists due to style requirements, so we
    # override it here if they are not set.
    FLAGS.hidden_layer_sizes = DEFAULT_LAYER_SIZES

  # Make the network.
  net = hk.without_apply_rng(hk.transform(net_fn))

  # Make the optimiser.
  opt = optax.adam(FLAGS.step_size)

  @jax.jit
  def loss(
      params: Params,
      inputs: np.ndarray,
      targets: np.ndarray,
  ) -> jax.Array:
    """Cross-entropy loss."""
    assert targets.dtype == np.int32
    log_probs = net.apply(params, inputs)
    return -jnp.mean(one_hot(targets, NUM_ACTIONS) * log_probs)

  @jax.jit
  def accuracy(
      params: Params,
      inputs: np.ndarray,
      targets: np.ndarray,
  ) -> jax.Array:
    """Classification accuracy."""
    predictions = net.apply(params, inputs)
    return jnp.mean(jnp.argmax(predictions, axis=-1) == targets)

  @jax.jit
  def update(
      params: Params,
      opt_state: OptState,
      inputs: np.ndarray,
      targets: np.ndarray,
  ) -> Tuple[Params, OptState]:
    """Learning rule (stochastic gradient descent)."""
    _, gradient = jax.value_and_grad(loss)(params, inputs, targets)
    updates, opt_state = opt.update(gradient, opt_state)
    new_params = optax.apply_updates(params, updates)
    return new_params, opt_state

  def output_samples(params: Params, max_samples: int):
    """Output some cases where the policy disagrees with the dataset action."""
    if max_samples == 0:
      return
    count = 0
    with open(os.path.join(FLAGS.data_path, 'test.txt')) as f:
      lines = list(f)
    np.random.shuffle(lines)
    for line in lines:
      state = GAME.new_initial_state()
      actions = _trajectory(line)
      for action in actions:
        if not state.is_chance_node():
          observation = np.array(state.information_state_tensor(), np.float32)
          policy = np.exp(net.apply(params, observation))
          probs_actions = [(p, a) for a, p in enumerate(policy)]
          pred = max(probs_actions)[1]
          if pred != action:
            print(state)
            for p, a in reversed(sorted(probs_actions)[-TOP_K_ACTIONS:]):
              print('{:7} {:.2f}'.format(state.action_to_string(a), p))
            print('Ground truth {}\n'.format(state.action_to_string(action)))
            count += 1
            break
        state.apply_action(action)
      if count >= max_samples:
        return

  # Store what we need to rebuild the Haiku net.
  if FLAGS.save_path:
    filename = os.path.join(FLAGS.save_path, 'layers.txt')
    with open(filename, 'w') as layer_def_file:
      for s in FLAGS.hidden_layer_sizes:
        layer_def_file.write(f'{s} ')
      layer_def_file.write('\n')

  # Make datasets.
  if FLAGS.data_path is None:
    raise app.UsageError(
        'Please generate your own supervised training data and supply the local'
        'location as --data_path')
  train = batch(
      make_dataset(os.path.join(FLAGS.data_path, 'train.txt')),
      FLAGS.train_batch)
  test = batch(
      make_dataset(os.path.join(FLAGS.data_path, 'test.txt')), FLAGS.eval_batch)

  # Initialize network and optimiser.
  if FLAGS.checkpoint_file:
    with open(FLAGS.checkpoint_file, 'rb') as pkl_file:
      params, opt_state = pickle.load(pkl_file)
  else:
    rng = jax.random.PRNGKey(FLAGS.rng_seed)  # seed used for network weights
    inputs, unused_targets = next(train)
    params = net.init(rng, inputs)
    opt_state = opt.init(params)

  # Train/eval loop.
  for step in range(FLAGS.iterations):
    # Do SGD on a batch of training examples.
    inputs, targets = next(train)
    params, opt_state = update(params, opt_state, inputs, targets)

    # Periodically evaluate classification accuracy on the test set.
    if (1 + step) % FLAGS.eval_every == 0:
      inputs, targets = next(test)
      test_accuracy = accuracy(params, inputs, targets)
      print(f'After {1+step} steps, test accuracy: {test_accuracy}.')
      if FLAGS.save_path:
        filename = os.path.join(FLAGS.save_path, f'checkpoint-{1 + step}.pkl')
        with open(filename, 'wb') as pkl_file:
          pickle.dump((params, opt_state), pkl_file)
      output_samples(params, FLAGS.num_examples)


if __name__ == '__main__':
  app.run(main)



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example: MMD with dilated entropy to compute approx. Nash in Kuhn poker."""

from absl import app
from absl import flags

from open_spiel.python.algorithms import exploitability
from open_spiel.python.algorithms import mmd_dilated
import pyspiel

FLAGS = flags.FLAGS

flags.DEFINE_integer("iterations", 1000, "Number of iterations")
flags.DEFINE_string("game", "kuhn_poker", "Name of the game")
flags.DEFINE_integer("print_freq", 100, "How often to print the exploitability")


def main(_):
  game = pyspiel.load_game(FLAGS.game)
  # need to manually set stepsize if alpha = 0
  mmd = mmd_dilated.MMDDilatedEnt(game, alpha=0, stepsize=1)

  for i in range(FLAGS.iterations):
    mmd.update_sequences()
    if i % FLAGS.print_freq == 0:
      conv = exploitability.exploitability(game, mmd.get_avg_policies())
      print("Iteration {} exploitability {}".format(i, conv))


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A cliff walking single agent reinforcement learning environment."""

import numpy as np

from open_spiel.python import rl_environment

# Actions
RIGHT, UP, LEFT, DOWN = range(4)


class Environment(object):
  r"""A cliff walking reinforcement learning environment.

  This is a deterministic environment that can be used to test RL algorithms.
  Note there are *no illegal moves* in this environment--if the agent is on the
  edge of the cliff and takes an action which would yield an invalid position,
  the action is ignored (as if there were walls surrounding the cliff).

  Cliff example for height=3 and width=5:

                |   |   |   |   |   |
                |   |   |   |   |   |
                | S | x | x | x | G |

  where `S` is always the starting position, `G` is always the goal and `x`
  represents the zone of high negative reward to be avoided. For this instance,
  the optimum policy is depicted as follows:

                |   |   |   |   |   |
                |-->|-->|-->|-->|\|/|
                |/|\| x | x | x | G |

  yielding a reward of -6 (minus 1 per time step).

  See pages 132 of Rich Sutton's book for details:
  http://www.incompleteideas.net/book/bookdraft2018mar21.pdf
  """

  def __init__(self, height=4, width=8, discount=1.0, max_t=100):
    if height < 2 or width < 3:
      raise ValueError("height must be >= 2 and width >= 3.")
    self._height = height
    self._width = width
    self._legal_actions = [RIGHT, UP, LEFT, DOWN]
    self._should_reset = True
    self._max_t = max_t

    # Discount returned at non-initial steps.
    self._discounts = [discount] * self.num_players

  def reset(self):
    """Resets the environment."""
    self._should_reset = False
    self._time_counter = 0
    self._state = np.array([self._height - 1, 0])

    observations = {
        "info_state": [self._state.copy()],
        "legal_actions": [self._legal_actions],
        "current_player": 0,
    }

    return rl_environment.TimeStep(
        observations=observations,
        rewards=None,
        discounts=None,
        step_type=rl_environment.StepType.FIRST)

  def step(self, actions):
    """Updates the environment according to `actions` and returns a `TimeStep`.

    Args:
      actions: A singleton list with an integer, or an integer, representing the
        action the agent took.

    Returns:
      A `rl_environment.TimeStep` namedtuple containing:
        observation: singleton list of dicts containing player observations,
            each corresponding to `observation_spec()`.
        reward: singleton list containing the reward at this timestep, or None
            if step_type is `rl_environment.StepType.FIRST`.
        discount: singleton list containing the discount in the range [0, 1], or
            None if step_type is `rl_environment.StepType.FIRST`.
        step_type: A `rl_environment.StepType` value.
    """
    if self._should_reset:
      return self.reset()
    self._time_counter += 1

    if isinstance(actions, list):
      action = actions[0]
    elif isinstance(actions, int):
      action = actions
    else:
      raise ValueError("Action not supported.", actions)

    dx = 0
    dy = 0
    if action == LEFT:
      dx -= 1
    elif action == RIGHT:
      dx += 1

    if action == UP:
      dy -= 1
    elif action == DOWN:
      dy += 1

    self._state += np.array([dy, dx])
    self._state = self._state.clip(0, [self._height - 1, self._width - 1])

    done = self._is_pit(self._state) or self._is_goal(self._state)
    done = done or self._time_counter >= self._max_t
    # Return observation
    step_type = (
        rl_environment.StepType.LAST if done else rl_environment.StepType.MID)
    self._should_reset = step_type == rl_environment.StepType.LAST

    observations = {
        "info_state": [self._state.copy()],
        "legal_actions": [self._legal_actions],
        "current_player": 0,
    }

    return rl_environment.TimeStep(
        observations=observations,
        rewards=[self._get_reward(self._state)],
        discounts=self._discounts,
        step_type=step_type)

  def _is_goal(self, pos):
    """Check if position is bottom right corner of grid."""
    return pos[0] == self._height - 1 and pos[1] == self._width - 1

  def _is_pit(self, pos):
    """Check if position is in bottom row between start and goal."""
    return (pos[1] > 0 and pos[1] < self._width - 1 and
            pos[0] == self._height - 1)

  def _get_reward(self, pos):
    if self._is_pit(pos):
      return -100.0
    else:
      return -1.0

  def observation_spec(self):
    """Defines the observation provided by the environment.

    Each dict member will contain its expected structure and shape.

    Returns:
      A specification dict describing the observation fields and shapes.
    """
    return dict(
        info_state=tuple([2]),
        legal_actions=(len(self._legal_actions),),
        current_player=(),
    )

  def action_spec(self):
    """Defines action specifications.

    Specifications include action boundaries and their data type.

    Returns:
      A specification dict containing action properties.
    """
    return dict(
        num_actions=len(self._legal_actions),
        min=min(self._legal_actions),
        max=max(self._legal_actions),
        dtype=int,
    )

  @property
  def num_players(self):
    return 1

  @property
  def is_turn_based(self):
    return False



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Python nfg_writer example."""

from absl import app
from absl import flags

import pyspiel

FLAGS = flags.FLAGS

flags.DEFINE_string("game", "matrix_rps", "Name of the game")
flags.DEFINE_string("outfile", None, "File to send the output to.")


def main(_):
  game = pyspiel.load_game(FLAGS.game)
  nfg_text = pyspiel.game_to_nfg_string(game)

  if FLAGS.outfile is None:
    print(nfg_text)
  else:
    print("Exporting to {}".format(FLAGS.outfile))
    outfile = open(FLAGS.outfile, "w")
    outfile.write(nfg_text)
    outfile.close()


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit test for XinXin MCTS bot."""

from absl.testing import absltest
import numpy as np

from open_spiel.python.algorithms import evaluate_bots
import pyspiel

SEED = 12983641


class ISMCTSBotTest(absltest.TestCase):

  def xinxin_play_game(self, game):
    bots = []
    for _ in range(4):
      bots.append(pyspiel.make_xinxin_bot(game.get_parameters()))
    evaluate_bots.evaluate_bots(game.new_initial_state(), bots, np.random)

  def test_basic_xinxin_selfplay(self):
    game = pyspiel.load_game("hearts")
    self.xinxin_play_game(game)


if __name__ == "__main__":
  np.random.seed(SEED)
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for open_spiel.python.algorithms.ppo."""

import random
from absl.testing import absltest
import numpy as np
import torch

from open_spiel.python import rl_environment
import pyspiel
from open_spiel.python.pytorch.ppo import PPO
from open_spiel.python.pytorch.ppo import PPOAgent
from open_spiel.python.vector_env import SyncVectorEnv

# A simple two-action game encoded as an EFG game. Going left gets -1, going
# right gets a +1.
SIMPLE_EFG_DATA = """
  EFG 2 R "Simple single-agent problem" { "Player 1" } ""
  p "ROOT" 1 1 "ROOT" { "L" "R" } 0
    t "L" 1 "Outcome L" { -1.0 }
    t "R" 2 "Outcome R" { 1.0 }
"""
SEED = 24261711


class PPOTest(absltest.TestCase):

  def test_simple_game(self):
    game = pyspiel.load_efg_game(SIMPLE_EFG_DATA)
    env = rl_environment.Environment(game=game)
    envs = SyncVectorEnv([env])
    agent_fn = PPOAgent
    anneal_lr = True

    info_state_shape = tuple(
        np.array(env.observation_spec()["info_state"]).flatten())

    total_timesteps = 1000
    steps_per_batch = 8
    batch_size = int(len(envs) * steps_per_batch)
    num_updates = total_timesteps // batch_size
    agent = PPO(
        input_shape=info_state_shape,
        num_actions=game.num_distinct_actions(),
        num_players=game.num_players(),
        player_id=0,
        num_envs=1,
        agent_fn=agent_fn,
    )

    time_step = envs.reset()
    for update in range(num_updates):
      for _ in range(steps_per_batch):
        agent_output = agent.step(time_step)
        time_step, reward, done, _ = envs.step(
            agent_output, reset_if_done=True)
        agent.post_step(reward, done)

      if anneal_lr:
        agent.anneal_learning_rate(update, num_updates)

      agent.learn(time_step)

    total_eval_reward = 0
    n_total_evaluations = 1000
    n_evaluations = 0
    time_step = envs.reset()
    while n_evaluations < n_total_evaluations:
      agent_output = agent.step(time_step, is_evaluation=True)
      time_step, reward, done, _ = envs.step(
          agent_output, reset_if_done=True)
      total_eval_reward += reward[0][0]
      n_evaluations += sum(done)
    self.assertGreaterEqual(total_eval_reward, 900)


if __name__ == "__main__":
  random.seed(SEED)
  torch.manual_seed(SEED)
  np.random.seed(SEED)
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Policy gradient methods implemented in JAX."""

import collections
import chex
import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np
import optax
import rlax

from open_spiel.python import rl_agent

Transition = collections.namedtuple(
    "Transition",
    "info_state action reward discount legal_actions_mask next_info_state")


class NetA2C(hk.Module):
  """A simple network with a policy head and a baseline value head."""

  def __init__(self, num_actions, hidden_layers_sizes):
    super().__init__()
    self._num_actions = num_actions
    self._hidden_layers_sizes = hidden_layers_sizes

  def __call__(self, info_state):
    """Process a batch of observations."""
    torso = hk.nets.MLP(self._hidden_layers_sizes, activate_final=True)
    hidden = torso(info_state)
    policy_logits = hk.Linear(self._num_actions)(hidden)
    baseline = hk.Linear(1)(hidden)
    return policy_logits, baseline


class NetPG(hk.Module):
  """A simple network with a policy head and an action-value head."""

  def __init__(self, num_actions, hidden_layers_sizes):
    super().__init__()
    self._num_actions = num_actions
    self._hidden_layers_sizes = hidden_layers_sizes

  def __call__(self, info_state):
    """Process a batch of observations."""
    torso = hk.nets.MLP(self._hidden_layers_sizes, activate_final=True)
    hidden = torso(info_state)
    policy_logits = hk.Linear(self._num_actions)(hidden)
    q_values = hk.Linear(self._num_actions)(hidden)
    return policy_logits, q_values


def generate_a2c_pi_loss(net_apply, loss_class, entropy_cost, l2_actor_weight,
                         lambda_):
  """A function generator generates loss function."""

  def _a2c_pi_loss(net_params, batch):
    info_states, actions, rewards, discounts = batch["info_states"], batch[
        "actions"], batch["rewards"], batch["discounts"]
    policy_logits, baselines = net_apply(net_params, info_states)
    policy_logits = policy_logits[:-1]

    baselines = jnp.squeeze(baselines, axis=1)
    baselines = jnp.concatenate([baselines[:-1], jnp.zeros(1)])
    td_returns = rlax.lambda_returns(
        rewards,
        discounts,
        baselines[1:],
        lambda_=lambda_,
        stop_target_gradients=True)
    advantages = td_returns - baselines[:-1]
    chex.assert_equal_shape([td_returns, actions, advantages])
    pi_loss = loss_class(
        logits_t=policy_logits,
        a_t=actions,
        adv_t=advantages,
        w_t=jnp.ones(td_returns.shape))
    ent_loss = rlax.entropy_loss(
        logits_t=policy_logits, w_t=jnp.ones(td_returns.shape))
    l2_loss = jnp.sum(jnp.square(jax.flatten_util.ravel_pytree(net_params)[0]))
    return pi_loss + entropy_cost * ent_loss + l2_actor_weight * l2_loss

  return _a2c_pi_loss


def generate_a2c_critic_loss(net_apply, l2_critic_weight, lambda_):
  """A function generator generates loss function."""

  def _a2c_critic_loss(net_params, batch):
    info_states, rewards, discounts = batch["info_states"], batch[
        "rewards"], batch["discounts"]
    _, baselines = net_apply(net_params, info_states)
    baselines = jnp.squeeze(baselines, axis=1)
    baselines = jnp.concatenate([baselines[:-1], jnp.zeros(1)])

    td_lambda = rlax.td_lambda(
        v_tm1=baselines[:-1],
        r_t=rewards,
        discount_t=discounts,
        v_t=baselines[1:],
        lambda_=lambda_,
        stop_target_gradients=True)
    l2_loss = jnp.sum(jnp.square(jax.flatten_util.ravel_pytree(net_params)[0]))
    return jnp.mean(jnp.square(td_lambda)) + l2_critic_weight * l2_loss

  return _a2c_critic_loss


def generate_pg_pi_loss(net_apply, loss_class, entropy_cost, l2_actor_weight):
  """A function generator generates loss function."""

  def _pg_loss(net_params, batch):
    info_states = batch["info_states"]
    policy_logits, q_values = net_apply(net_params, info_states)
    chex.assert_equal_shape([policy_logits, q_values])
    pi_loss = loss_class(logits_t=policy_logits, q_t=q_values)
    ent_loss = rlax.entropy_loss(
        logits_t=policy_logits, w_t=jnp.ones(policy_logits.shape[:1]))
    l2_loss = jnp.sum(jnp.square(jax.flatten_util.ravel_pytree(net_params)[0]))
    return pi_loss + entropy_cost * ent_loss + l2_actor_weight * l2_loss

  return _pg_loss


def generate_pg_critic_loss(net_apply, l2_critic_weight, lambda_):
  """A function generator generates loss function."""

  def _critic_loss(net_params, batch):
    info_states, actions, rewards, discounts = batch["info_states"], batch[
        "actions"], batch["rewards"], batch["discounts"]
    _, q_values = net_apply(net_params, info_states)
    q_values = q_values[:-1]
    q_values = jnp.concatenate(
        [q_values, jnp.zeros(q_values[-1].reshape(1, -1).shape)])

    actions = jnp.concatenate([actions, jnp.zeros(1, dtype=int)])
    sarsa_lambda = rlax.sarsa_lambda(
        q_tm1=q_values[:-1],
        a_tm1=actions[:-1],
        r_t=rewards,
        discount_t=discounts,
        q_t=q_values[1:],
        a_t=actions[1:],
        lambda_=lambda_,
        stop_target_gradients=True)
    l2_loss = jnp.sum(jnp.square(jax.flatten_util.ravel_pytree(net_params)[0]))
    return jnp.mean(jnp.square(sarsa_lambda)) + l2_critic_weight * l2_loss

  return _critic_loss


def generate_act_func(net_apply):
  """A function generator generates act function."""

  def _act(net_params, info_state, action_mask, rng):
    info_state = jnp.reshape(info_state, [1, -1])
    policy_logits, _ = net_apply(net_params, info_state)
    policy_probs = jax.nn.softmax(policy_logits, axis=1)

    # Remove illegal actions, re-normalize probs
    probs = policy_probs[0] * action_mask

    probs /= jnp.sum(probs)
    action = jax.random.choice(rng, len(probs), p=probs)
    return action, probs

  return _act


class PolicyGradient(rl_agent.AbstractAgent):
  """Policy Gradient Agent implementation in JAX."""

  def __init__(self,
               player_id,
               info_state_size,
               num_actions,
               loss_str="a2c",
               loss_class=None,
               hidden_layers_sizes=(128,),
               lambda_=1.0,
               critic_learning_rate=0.01,
               pi_learning_rate=0.001,
               entropy_cost=0.01,
               l2_weight_actor=0.0,
               l2_weight_critic=0.0,
               num_critic_before_pi=8,
               additional_discount_factor=1.0,
               max_global_gradient_norm=None,
               optimizer_str="sgd",
               seed=42):
    """Initialize the PolicyGradient agent.

    Args:
      player_id: int, player identifier. Usually its position in the game.
      info_state_size: int, info_state vector size.
      num_actions: int, number of actions per info state.
      loss_str: string or None. If string, must be one of ["rpg", "qpg", "rm",
        "a2c"] and defined in `_get_loss_class`. If None, a loss class must be
        passed through `loss_class`. Defaults to "a2c".
      loss_class: Class or None. If Class, it must define the policy gradient
        loss. If None a loss class in a string format must be passed through
        `loss_str`. Defaults to None.
      hidden_layers_sizes: iterable, defines the neural network layers. Defaults
        to (128,), which produces a NN: [INPUT] -> [128] -> ReLU -> [OUTPUT].
      lambda_: float, lambda in TD(lambda) or SARSA(lambda). Defaults to 1.0.
      critic_learning_rate: float, learning rate used for Critic (Q or V).
        Defaults to 0.001.
      pi_learning_rate: float, learning rate used for Pi. Defaults to 0.001.
      entropy_cost: float, entropy cost used to multiply the entropy loss. Can
        be set to None to skip entropy computation. Defaults to 0.001.
      l2_weight_actor: l2 penaly weight for actor network. Defaults to 0.0.
      l2_weight_critic: l2 penalty weight for critic network. Defaults to
        0.0.
      num_critic_before_pi: int, number of Critic (Q or V) updates before each
        Pi update. Defaults to 8 (every 8th critic learning step, Pi also
        learns).
      additional_discount_factor: float, additional discount to compute returns.
        Defaults to 1.0, in which case, no extra discount is applied.  None that
        users must provide *only one of* `loss_str` or `loss_class`.
      max_global_gradient_norm: float or None, maximum global norm of a gradient
        to which the gradient is shrunk if its value is larger.
      optimizer_str: String defining which optimizer to use. Supported values
        are {sgd, adam}
      seed: random seed
    """
    assert bool(loss_str) ^ bool(loss_class), "Please provide only one option."
    self._kwargs = locals()
    loss_class = loss_class if loss_class else self._get_loss_class(loss_str)

    self.player_id = player_id
    self._num_actions = num_actions
    self._extra_discount = additional_discount_factor
    self._num_critic_before_pi = num_critic_before_pi

    self._episode_data = []
    self._dataset = collections.defaultdict(list)
    self._prev_time_step = None
    self._prev_action = None

    # Step counters
    self._step_counter = 0
    self._episode_counter = 0
    self._num_learn_steps = 0

    # Keep track of the last training loss achieved in an update step.
    self._last_loss_value = None

    self._loss_str = loss_str

    # Network
    # activate final as we plug logit and qvalue heads afterwards.
    net_class = NetA2C if loss_str == "a2c" else NetPG

    def net_func(info_input):
      net = net_class(num_actions, hidden_layers_sizes)
      return net(info_input)

    hk_net = hk.without_apply_rng(hk.transform(net_func))

    hk_net_apply = hk_net.apply
    self.rng = jax.random.PRNGKey(seed)
    init_inputs = jnp.ones((1, info_state_size))
    self.hk_net_params = hk_net.init(self.rng, init_inputs)

    self._act = jax.jit(generate_act_func(hk_net_apply))

    if optimizer_str == "adam":
      critic_optimizer = optax.adam(critic_learning_rate)
      pi_optimizer = optax.adam(pi_learning_rate)

    elif optimizer_str == "sgd":
      critic_optimizer = optax.sgd(critic_learning_rate)
      pi_optimizer = optax.sgd(pi_learning_rate)

    else:
      raise ValueError("Not implemented, choose from 'adam' and 'sgd'.")

    if max_global_gradient_norm:
      pi_optimizer = optax.chain(
          pi_optimizer, optax.clip_by_global_norm(max_global_gradient_norm))
      critic_optimizer = optax.chain(
          critic_optimizer, optax.clip_by_global_norm(max_global_gradient_norm))

    pi_opt_init, pi_opt_update = pi_optimizer.init, pi_optimizer.update
    critic_opt_init, critic_opt_update = critic_optimizer.init, critic_optimizer.update

    self._pi_opt_state = pi_opt_init(self.hk_net_params)

    if loss_str == "a2c":
      pi_loss_and_grad = jax.value_and_grad(
          generate_a2c_pi_loss(hk_net_apply, loss_class, entropy_cost,
                               l2_weight_actor, lambda_))
      critic_loss_and_grad = jax.value_and_grad(
          generate_a2c_critic_loss(hk_net_apply, l2_weight_critic, lambda_))
      self._critic_opt_state = critic_opt_init(self.hk_net_params)
    else:
      pi_loss_and_grad = jax.value_and_grad(
          generate_pg_pi_loss(hk_net_apply, loss_class, entropy_cost,
                              l2_weight_actor))
      critic_loss_and_grad = jax.value_and_grad(
          generate_pg_critic_loss(hk_net_apply, l2_weight_critic, lambda_))
      self._critic_opt_state = critic_opt_init(self.hk_net_params)

    self._jit_pi_update = jax.jit(
        self._get_update(pi_opt_update, pi_loss_and_grad))
    self._jit_critic_update = jax.jit(
        self._get_update(critic_opt_update, critic_loss_and_grad))

  def _get_loss_class(self, loss_str):
    if loss_str == "rpg":
      return rlax.rpg_loss
    elif loss_str == "qpg":
      return rlax.qpg_loss
    elif loss_str == "rm":
      return rlax.rm_loss
    elif loss_str == "a2c":
      return rlax.policy_gradient_loss

  def _get_update(self, opt_update, loss_fn):

    def update(net_params, opt_state, batch):
      loss_val, grad_val = loss_fn(net_params, batch)
      updates, new_opt_state = opt_update(grad_val, opt_state)
      new_net_params = optax.apply_updates(net_params, updates)
      return new_net_params, new_opt_state, loss_val

    return update

  def step(self, time_step, is_evaluation=False):
    """Returns the action to be taken and updates the network if needed.

    Args:
      time_step: an instance of rl_environment.TimeStep.
      is_evaluation: bool, whether this is a training or evaluation call.

    Returns:
      A `rl_agent.StepOutput` containing the action probs and chosen action.
    """
    # Act step: don't act at terminal info states or if its not our turn.
    if (not time_step.last()) and (time_step.is_simultaneous_move() or
                                   self.player_id
                                   == time_step.current_player()):
      info_state = time_step.observations["info_state"][self.player_id]
      legal_actions = time_step.observations["legal_actions"][self.player_id]
      action_mask = np.zeros(self._num_actions)
      action_mask[legal_actions] = 1
      self.rng, _ = jax.random.split(self.rng)
      action, probs = self._act(self.hk_net_params, np.asarray(info_state),
                                action_mask, self.rng)
    else:
      action = None
      probs = []

    if not is_evaluation:
      self._step_counter += 1

      # Add data points to current episode buffer.
      if self._prev_time_step:
        self._add_transition(time_step)

      # Episode done, add to dataset and maybe learn.

      if time_step.last():
        self._episode_counter += 1

        self._critic_update()
        self._num_learn_steps += 1
        if self._num_learn_steps % self._num_critic_before_pi == 0:
          self._pi_update()
        self._episode_data = []

        self._prev_time_step = None
        self._prev_action = None
        return
      else:
        self._prev_time_step = time_step
        self._prev_action = action

    return rl_agent.StepOutput(action=action, probs=probs)

  @property
  def loss(self):
    return (self._last_critic_loss_value, self._last_pi_loss_value)

  def _add_transition(self, time_step):
    """Adds intra-episode transition to the `_episode_data` buffer.

    Adds the transition from `self._prev_time_step` to `time_step`.
    Args:
      time_step: an instance of rl_environment.TimeStep.
    """
    assert self._prev_time_step is not None
    legal_actions = (
        self._prev_time_step.observations["legal_actions"][self.player_id])
    legal_actions_mask = np.zeros(self._num_actions)
    legal_actions_mask[legal_actions] = 1.0
    transition = Transition(
        info_state=(
            self._prev_time_step.observations["info_state"][self.player_id][:]),
        action=self._prev_action,
        reward=time_step.rewards[self.player_id],
        discount=time_step.discounts[self.player_id],
        legal_actions_mask=legal_actions_mask,
        next_info_state=(
            time_step.observations["info_state"][self.player_id][:]))

    self._episode_data.append(transition)

  def _critic_update(self):
    """Compute the Critic loss on sampled transitions & perform a critic update.

    Returns:
      The average Critic loss obtained on this batch.
    """
    batch = {}
    batch["info_states"] = jnp.asarray(
        [transition.info_state for transition in self._episode_data] +
        [self._episode_data[-1].next_info_state])
    batch["rewards"] = jnp.asarray(
        [transition.reward for transition in self._episode_data])
    batch["discounts"] = jnp.asarray(
        [transition.discount for transition in self._episode_data])
    if self._loss_str != "a2c":
      batch["actions"] = jnp.asarray(
          [transition.action for transition in self._episode_data])

    self.hk_net_params, self._critic_opt_state, self._last_critic_loss_value = self._jit_critic_update(
        self.hk_net_params, self._critic_opt_state, batch)
    return self._last_critic_loss_value

  def _pi_update(self):
    """Compute the Pi loss on sampled transitions and perform a Pi update.

    Returns:
      The average Pi loss obtained on this batch.
    """
    batch = {}
    batch["info_states"] = jnp.asarray(
        [transition.info_state for transition in self._episode_data] +
        [self._episode_data[-1].next_info_state])

    if self._loss_str == "a2c":
      batch["discounts"] = jnp.asarray(
          [transition.discount for transition in self._episode_data])
      batch["actions"] = jnp.asarray(
          [transition.action for transition in self._episode_data])
      batch["rewards"] = jnp.asarray(
          [transition.reward for transition in self._episode_data])
    self.hk_net_params, self._pi_opt_state, self._last_pi_loss_value = self._jit_pi_update(
        self.hk_net_params, self._pi_opt_state, batch)
    return self._last_pi_loss_value



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""File utilities."""

import os


def find_file(filename, levels):
  if os.path.isfile(filename):
    return filename
  else:
    for _ in range(levels):
      filename = '../' + filename
      if os.path.isfile(filename):
        return filename
  return None




# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Value iteration algorithm for solving a game."""

from open_spiel.python.algorithms import get_all_states
from open_spiel.python.algorithms import lp_solver
import pyspiel


def _get_future_states(possibilities, state, reach=1.0):
  """Does a lookahead over chance nodes to all next states after (s,a).

  Also works if there are no chance nodes (i.e. base case).

  Arguments:
    possibilities:  an empty list, that will be filled with (str(next_state),
      transition probability) pairs for all possible next states
    state: the state following some s.apply_action(a), can be a chance node
    reach: chance reach probability of getting to this point from (s,a)
  Returns: nothing.
  """
  if not state.is_chance_node() or state.is_terminal():
    # Base case
    possibilities.append((str(state), reach))
  else:
    assert state.is_chance_node()
    for outcome, prob in state.chance_outcomes():
      next_state = state.child(outcome)
      _get_future_states(possibilities, next_state, reach * prob)


def _add_transition(transitions, key, state):
  """Adds action transitions from given state."""

  if state.is_simultaneous_node():
    for p0action in state.legal_actions(0):
      for p1action in state.legal_actions(1):
        next_state = state.clone()
        next_state.apply_actions([p0action, p1action])
        possibilities = []
        _get_future_states(possibilities, next_state)
        transitions[(key, p0action, p1action)] = possibilities
  else:
    for action in state.legal_actions():
      next_state = state.child(action)
      possibilities = []
      _get_future_states(possibilities, next_state)
      transitions[(key, action)] = possibilities


def _initialize_maps(states, values, transitions):
  """Initialize the value and transition maps."""
  for key, state in states.items():
    if state.is_terminal():
      values[key] = state.player_return(0)
    else:
      values[key] = 0
      _add_transition(transitions, key, state)


def value_iteration(game, depth_limit, threshold, cyclic_game=False):
  """Solves for the optimal value function of a game.

  For small games only! Solves the game using value iteration,
  with the maximum error for the value function less than threshold.
  This algorithm works for sequential 1-player games or 2-player zero-sum
  games, with or without chance nodes.

  Arguments:
    game: The game to analyze, as returned by `load_game`.
    depth_limit: How deeply to analyze the game tree. Negative means no limit, 0
      means root-only, etc.
    threshold: Maximum error for state values..
    cyclic_game: set to True if the game has cycles (from state A we can get to
      state B, and from state B we can get back to state A).

  Returns:
    A `dict` with string keys and float values, mapping string encoding of
    states to the values of those states.
  """
  assert game.num_players() in (1,
                                2), ("Game must be a 1-player or 2-player game")
  if game.num_players() == 2:
    assert game.get_type().utility == pyspiel.GameType.Utility.ZERO_SUM, (
        "2-player games must be zero sum games")

  # Must be perfect information or one-shot (not imperfect information).
  assert (game.get_type().information == pyspiel.GameType.Information.ONE_SHOT
          or game.get_type().information ==
          pyspiel.GameType.Information.PERFECT_INFORMATION)

  # We expect Value Iteration to be used with perfect information games, in
  # which `str` is assumed to display the state of the game.
  states = get_all_states.get_all_states(
      game,
      depth_limit,
      True,
      False,
      to_string=str,
      stop_if_encountered=cyclic_game)
  values = {}
  transitions = {}

  _initialize_maps(states, values, transitions)
  error = threshold + 1  # A value larger than threshold
  min_utility = game.min_utility()
  while error > threshold:
    error = 0
    for key, state in states.items():
      if state.is_terminal():
        continue
      elif state.is_simultaneous_node():
        # Simultaneous node. Assemble a matrix game from the child utilities.
        # and solve it using a matrix game solver.
        p0_utils = []  # row player
        p1_utils = []  # col player
        row = 0
        for p0action in state.legal_actions(0):
          # new row
          p0_utils.append([])
          p1_utils.append([])
          for p1action in state.legal_actions(1):
            # loop from left-to-right of columns
            next_states = transitions[(key, p0action, p1action)]
            joint_q_value = sum(
                p * values[next_state] for next_state, p in next_states)
            p0_utils[row].append(joint_q_value)
            p1_utils[row].append(-joint_q_value)
          row += 1
        stage_game = pyspiel.create_matrix_game(p0_utils, p1_utils)
        solution = lp_solver.solve_zero_sum_matrix_game(stage_game)
        value = solution[2]
      else:
        # Regular decision node
        player = state.current_player()
        value = min_utility if player == 0 else -min_utility
        for action in state.legal_actions():
          next_states = transitions[(key, action)]
          q_value = sum(p * values[next_state] for next_state, p in next_states)
          if player == 0:
            value = max(value, q_value)
          else:
            value = min(value, q_value)
      error = max(abs(values[key] - value), error)
      values[key] = value

  return values



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for open_spiel.python.algorithms.nash_averaging."""

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from open_spiel.python.algorithms.nash_averaging import nash_averaging
import pyspiel

# transitive game test case
game_trans = pyspiel.create_matrix_game(
    [[0.0, -1.0, -1.0], [1.0, 0.0, -1.0], [1.0, 1.0, 0.0]],
    [[0.0, 1.0, 1.0], [-1.0, 0.0, 1.0], [-1.0, -1.0, 0.0]])

eq_trans = np.asarray([0., 0., 1.])
value_trans = np.asarray([-1., -1., 0.])

# rock-paper-scissors test case
game_rps = pyspiel.create_matrix_game(
    [[0.0, -1.0, 1.0], [1.0, 0.0, -1.0], [-1.0, 1.0, 0.0]],
    [[0.0, 1.0, -1.0], [-1.0, 0.0, 1.0], [1.0, -1.0, 0.0]])
eq_rps = np.asarray([1 / 3, 1 / 3, 1 / 3])
value_rps = np.asarray([0., 0., 0.])

# game with one dominated strategy (AvA case)
p_mat0 = np.asarray([[0.0, 234., 34., -270.], [-234., 0., -38., -464.],
                     [-34., 38., 0., -270.], [270., 464., 270., 0.]])
game0 = pyspiel.create_matrix_game(p_mat0, -p_mat0)
dominated_idxs0 = [0, 1, 2]


# game with one dominated strategy (AvT case)
p_mat1 = np.asarray([
    [0.0, 0.0, 0.0],
    [1.0, 10.0, 100.0],
    [2.0, 20.0, 200.0],
    [3.0, 30.0, 300.0],
])
game1 = pyspiel.create_matrix_game(p_mat1, -p_mat1)
dominated_idxs1 = [0, 1, 2]


# game with one multiple dominant strategy (AvT case)
p_mat2 = np.asarray([
    [0.0, 0.0, 0.0],
    [1.0, 10.0, 100.0],
    [2.0, 20.0, 200.0],
    [3.0, 30.0, 300.0],
    [3.0, 30.0, 300.0],
])
game2 = pyspiel.create_matrix_game(p_mat2, -p_mat2)
dom_idxs2 = [3, 4]


class NashAveragingTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ("transitive_game", game_trans, eq_trans, value_trans),
      ("rps_game", game_rps, eq_rps, value_rps),
  )
  def test_simple_games(self, game, eq, value):

    maxent_nash, nash_avg_value = nash_averaging(game)
    with self.subTest("probability"):
      np.testing.assert_array_almost_equal(eq, maxent_nash.reshape(-1))

    with self.subTest("value"):
      np.testing.assert_array_almost_equal(value, nash_avg_value.reshape(-1))

  @parameterized.named_parameters(
      ("game0", game0, dominated_idxs0),)
  def test_ava_games_with_dominated_strategy(self, game, dominated_idxs):
    maxent_nash, _ = nash_averaging(game)
    with self.subTest("dominated strategies have zero Nash probs"):
      for idx in dominated_idxs:
        self.assertAlmostEqual(maxent_nash[idx].item(), 0.0)

  @parameterized.named_parameters(
      ("game1", game1, dominated_idxs1),
  )
  def test_avt_games_with_dominated_strategy(self, game, dominated_idxs):
    (agent_strategy, _), _ = nash_averaging(game, a_v_a=False)
    with self.subTest("dominated strategies have zero Nash probs"):
      for idx in dominated_idxs:
        self.assertAlmostEqual(agent_strategy[idx].item(), 0.0)

  @parameterized.named_parameters(
      ("game2", game2, dom_idxs2),
  )
  def test_avt_games_with_multiple_dominant_strategies(self, game, dom_idxs):
    (agent_strategy, _), (agent_values, _) = nash_averaging(game, a_v_a=False)
    with self.subTest("dominant strategies have equal Nash probs"):
      for idx in dom_idxs:
        self.assertAlmostEqual(agent_strategy[idx].item(), 1 / len(dom_idxs2))

    with self.subTest("dominant strategies have equal Nash values"):
      values = [agent_values[idx] for idx in dom_idxs]
      self.assertAlmostEqual(np.abs(np.max(values) - np.min(values)), 0.0)


if __name__ == "__main__":
  absltest.main()



"""Tests for google3.third_party.pybind11_abseil.status_casters."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import absltest
from pybind11_abseil import status
from pybind11_abseil.tests import status_example


def docstring_signature(f):
  """Returns the first line from a docstring - the signature for a function."""
  return f.__doc__.split('\n')[0]


class StatusTest(absltest.TestCase):

  def test_pass_status(self):
    test_status = status.Status(status.StatusCode.CANCELLED, 'test')
    self.assertTrue(
        status_example.check_status(test_status, status.StatusCode.CANCELLED))

  def test_return_status_return_type_from_doc(self):
    self.assertEndsWith(
        docstring_signature(status_example.return_status), ' -> None')

  def test_return_ok(self):
    # The return_status function should convert an ok status to None.
    self.assertIsNone(status_example.return_status(status.StatusCode.OK))

  def test_return_not_ok(self):
    # The return_status function should convert a non-ok status to an exception.
    with self.assertRaises(status.StatusNotOk) as cm:
      status_example.return_status(status.StatusCode.CANCELLED, 'test')
    self.assertEqual(cm.exception.status.code(), status.StatusCode.CANCELLED)
    self.assertEqual(cm.exception.status.message(), 'test')
    self.assertEqual(cm.exception.code, int(status.StatusCode.CANCELLED))
    self.assertEqual(cm.exception.message, 'test')

  def test_build_status_not_ok_enum(self):
    e = status.BuildStatusNotOk(status.StatusCode.INVALID_ARGUMENT, 'Msg enum.')
    self.assertEqual(e.status.code(), status.StatusCode.INVALID_ARGUMENT)
    self.assertEqual(e.code, int(status.StatusCode.INVALID_ARGUMENT))
    self.assertEqual(e.message, 'Msg enum.')

  def test_build_status_not_ok_int(self):
    e = status.BuildStatusNotOk(
        int(status.StatusCode.ALREADY_EXISTS), 'Msg int.')
    self.assertEqual(e.status.code(), status.StatusCode.ALREADY_EXISTS)
    self.assertEqual(e.code, int(status.StatusCode.ALREADY_EXISTS))
    self.assertEqual(e.message, 'Msg int.')

  def test_return_not_ok_twice(self):
    # Each exception is a different instance with different messages.
    with self.assertRaises(status.StatusNotOk) as cm1:
      status_example.return_status(status.StatusCode.CANCELLED, 'test1')
    with self.assertRaises(status.StatusNotOk) as cm2:
      status_example.return_status(status.StatusCode.CANCELLED, 'test2')
    self.assertEqual(cm1.exception.status.message(), 'test1')
    self.assertEqual(cm2.exception.status.message(), 'test2')

  def test_return_not_ok_catch_with_alias(self):
    # Catch as status_example.StatusNotOk, an alias of status.StatusNotOk.
    with self.assertRaises(status_example.StatusNotOk) as cm:
      status_example.return_status(status.StatusCode.CANCELLED, 'test')
    self.assertEqual(cm.exception.status.code(), status.StatusCode.CANCELLED)
    self.assertEqual(cm.exception.status.message(), 'test')

  def test_return_not_ok_catch_as_generic_exception(self):
    # Catch as a generic Exception, the base type of StatusNotOk.
    with self.assertRaises(Exception):
      status_example.return_status(status.StatusCode.CANCELLED, 'test')

  def test_make_status_return_type_from_doc(self):
    self.assertRegex(
        docstring_signature(status_example.make_status), r' -> .*\.Status')

  def test_make_ok(self):
    # The make_status function has been set up to return a status object
    # instead of raising an exception (this is done in status_example.cc).
    test_status = status_example.make_status(status.StatusCode.OK)
    self.assertTrue(test_status.ok())
    self.assertEqual(test_status.code(), status.StatusCode.OK)
    self.assertEqual(test_status.code_int(), 0)

  def test_make_not_ok(self):
    # The make_status function should always return a status object, even if
    # it is not ok (ie, it should *not* convert it to an exception).
    test_status = status_example.make_status(status.StatusCode.CANCELLED)
    self.assertFalse(test_status.ok())
    self.assertEqual(test_status.code(), status.StatusCode.CANCELLED)
    self.assertEqual(test_status.code_int(), 1)

  def test_make_not_ok_manual_cast(self):
    test_status = status_example.make_status_manual_cast(
        status.StatusCode.CANCELLED)
    self.assertEqual(test_status.code(), status.StatusCode.CANCELLED)

  def test_make_status_ref(self):
    result_1 = status_example.make_status_ref(status.StatusCode.OK)
    self.assertEqual(result_1.code(), status.StatusCode.OK)
    result_2 = status_example.make_status_ref(status.StatusCode.CANCELLED)
    self.assertEqual(result_2.code(), status.StatusCode.CANCELLED)
    # result_1 and 2 reference the same value, so they should always be equal.
    self.assertEqual(result_1.code(), result_2.code())

  def test_make_status_ptr(self):
    result_1 = status_example.make_status_ptr(status.StatusCode.OK)
    self.assertEqual(result_1.code(), status.StatusCode.OK)
    result_2 = status_example.make_status_ptr(status.StatusCode.CANCELLED)
    self.assertEqual(result_2.code(), status.StatusCode.CANCELLED)
    # result_1 and 2 reference the same value, so they should always be equal.
    self.assertEqual(result_1.code(), result_2.code())

  def test_canonical_error(self):
    test_status = status.aborted_error('test')
    self.assertEqual(test_status.code(), status.StatusCode.ABORTED)
    self.assertEqual(test_status.message(), 'test')

  def test_member_method(self):
    test_status = status_example.TestClass().make_status(status.StatusCode.OK)
    self.assertEqual(test_status.code(), status.StatusCode.OK)
    test_status = status_example.TestClass().make_status_const(
        status.StatusCode.OK)
    self.assertEqual(test_status.code(), status.StatusCode.OK)

  def test_is_ok(self):
    ok_status = status_example.make_status(status.StatusCode.OK)
    self.assertTrue(status.is_ok(ok_status))
    failure_status = status_example.make_status(status.StatusCode.CANCELLED)
    self.assertFalse(status.is_ok(failure_status))

  def test_ok_to_string(self):
    ok_status = status_example.make_status(status.StatusCode.OK)
    self.assertEqual(ok_status.to_string(), 'OK')
    self.assertEqual(repr(ok_status), 'OK')
    self.assertEqual(str(ok_status), 'OK')

  def test_canonical_error_to_string(self):
    test_status = status.aborted_error('test')
    self.assertEqual(test_status.to_string(), 'ABORTED: test')
    self.assertEqual(repr(test_status), 'ABORTED: test')
    self.assertEqual(str(test_status), 'ABORTED: test')

  def test_create_ok_status(self):
    ok_status = status.Status.OkStatus()
    self.assertEqual(ok_status.to_string(), 'OK')
    self.assertEqual(ok_status.raw_code(), 0)
    self.assertEqual(ok_status.CanonicalCode(), 0)
    self.assertEqual(ok_status.error_message(), '')
    self.assertIsNone(ok_status.IgnoreError())

  def test_raw_code_ne_code(self):
    st500 = status_example.status_from_int_code(500, 'Not a canonical code.')
    self.assertEqual(st500.raw_code(), 500)
    self.assertEqual(st500.code(), status.StatusCode.UNKNOWN)


class IntGetter(status_example.IntGetter):

  def Get(self, i):
    if i > 10:
      raise ValueError('Value out of range')
    return i


class StatusOrTest(absltest.TestCase):

  def test_return_value_status_or_return_type_from_doc(self):
    self.assertEndsWith(
        docstring_signature(status_example.return_value_status_or),
        ' -> int')

  def test_return_value(self):
    self.assertEqual(status_example.return_value_status_or(5), 5)

  def test_return_not_ok(self):
    with self.assertRaises(status.StatusNotOk) as cm:
      status_example.return_failure_status_or(status.StatusCode.NOT_FOUND)
    self.assertEqual(cm.exception.status.code(), status.StatusCode.NOT_FOUND)

  def test_make_failure_status_or_return_type_from_doc(self):
    self.assertRegex(
        docstring_signature(status_example.make_failure_status_or),
        r' -> Union\[.*\.Status, int\]')

  def test_make_not_ok(self):
    self.assertEqual(
        status_example.make_failure_status_or(
            status.StatusCode.CANCELLED).code(), status.StatusCode.CANCELLED)

  def test_make_not_ok_manual_cast(self):
    self.assertEqual(
        status_example.make_failure_status_or_manual_cast(
            status.StatusCode.CANCELLED).code(), status.StatusCode.CANCELLED)

  def test_return_ptr_status_or(self):
    result_1 = status_example.return_ptr_status_or(5)
    self.assertEqual(result_1.value, 5)
    result_2 = status_example.return_ptr_status_or(6)
    self.assertEqual(result_2.value, 6)
    # result_1 and 2 reference the same value, so they should always be equal.
    self.assertEqual(result_1.value, result_2.value)

  def test_return_unique_ptr(self):
    result = status_example.return_unique_ptr_status_or(5)
    self.assertEqual(result.value, 5)

  def test_member_method(self):
    test_status = status_example.TestClass().make_failure_status_or(
        status.StatusCode.ABORTED)
    self.assertEqual(test_status.code(), status.StatusCode.ABORTED)

  def test_is_ok(self):
    ok_result = status_example.return_value_status_or(5)
    self.assertEqual(ok_result, 5)
    self.assertTrue(status.is_ok(ok_result))
    failure_result = status_example.make_failure_status_or(
        status.StatusCode.CANCELLED)
    self.assertFalse(status.is_ok(failure_result))

  def test_return_status_or_pointer(self):
    expected_result = 42
    for _ in range(3):
      result = status_example.return_status_or_pointer()
      self.assertEqual(result, expected_result)

  def test_return_failed_status_or_pointer(self):
    for _ in range(3):
      with self.assertRaises(status.StatusNotOk):
        status_example.return_failure_status_or_pointer()

  def test_canonical_error_to_string(self):
    failure_result = status_example.make_failure_status_or(
        status.StatusCode.CANCELLED)
    self.assertEqual(failure_result.to_string(), 'CANCELLED: ')
    self.assertEqual(repr(failure_result), 'CANCELLED: ')
    self.assertEqual(str(failure_result), 'CANCELLED: ')

  def test_overriding_in_python(self):
    int_getter = IntGetter()
    self.assertEqual(int_getter.Get(5), 5)
    with self.assertRaises(ValueError):
      int_getter.Get(100)


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Big tensor games."""

from absl import logging  # pylint:disable=unused-import

import numpy as np

from open_spiel.python.algorithms.adidas_utils.helpers import misc


class TensorGame(object):
  """Tensor Game."""

  def __init__(self, pt, seed=None):
    """Ctor. Inits payoff tensor (players x actions x ... np.array).

    Args:
      pt: payoff tensor, np.array
      seed: seed for random number generator, used if computing best responses
    """
    if np.any(pt < 0.):
      raise ValueError("Payoff tensor must contain non-negative values")
    self.pt = pt

    self.seed = seed
    self.random = np.random.RandomState(seed)

  def num_players(self):
    return self.pt.shape[0]

  def num_strategies(self):
    return self.pt.shape[1:]

  def payoff_tensor(self):
    return self.pt

  def get_payoffs_for_strategies(self, policies):
    """Return vector of payoffs for all players given list of strategies.

    Args:
      policies: list of integers indexing strategies for each player
    Returns:
      np.array (length num players) of payoffs
    """
    return self.pt[tuple([slice(None)] + policies)]

  def best_response(self, mixed_strategy, return_exp=False):
    """Return best response and its superiority over the current strategy.

    Args:
      mixed_strategy: np.ndarray (distribution over strategies)
      return_exp: bool, whether to return how much best response exploits the
        given mixed strategy (default is False)
    Returns:
      br: int, index of strategy (ties split randomly)
      exp: u(br) - u(mixed_strategy)
    """
    logging.warn("Assumes symmetric game! Returns br for player 0.")
    gradient = misc.pt_reduce(self.pt[0],
                              [mixed_strategy] * self.num_players(),
                              [0])
    br = misc.argmax(self.random, gradient)
    exp = gradient.max() - gradient.dot(mixed_strategy)
    if return_exp:
      return br, exp
    else:
      return br

  def best_population_response(self, dist, policies):
    """Returns the best response to the current population of policies.

    Args:
      dist: np.ndarray, distribution over policies
      policies: list of integers indexing strategies for each player
    Returns:
      best response, exploitability tuple (see best_response)
    """
    ns = self.num_strategies()
    mixed_strat = np.zeros(ns)
    for pure_strat, prob in zip(policies, dist):
      mixed_strat[pure_strat] += prob
    return self.best_response(mixed_strat)


class ElFarol(TensorGame):
  """N-Player, 2-Action symmetric game with unique symmetric Nash."""

  def __init__(self, n=2, c=0.5, B=0, S=1, G=2, seed=None):
    """Ctor. Initializes payoff tensor (N x (2,) * N np.array).

    See Section 3.1, The El Farol Stage Game in
    http://www.econ.ed.ac.uk/papers/id186_esedps.pdf

    action 0: go to bar
    action 1: avoid bar

    Args:
      n: int, number of players
      c: float, threshold for `crowded' as a fraction of number of players
      B: float, payoff for going to a crowded bar
      S: float, payoff for staying at home
      G: float, payoff for going to an uncrowded bar
      seed: seed for random number generator, used if computing best responses
    """
    assert G > S > B, "Game parameters must satisfy G > S > B."
    pt = np.zeros((n,) + (2,) * n)
    for idx in np.ndindex(pt.shape):
      p = idx[0]
      a = idx[1:]
      a_i = a[p]
      go_to_bar = (a_i < 1)
      crowded = (n - 1 - sum(a) + a_i) >= (c * n)
      if go_to_bar and not crowded:
        pt[idx] = G
      elif go_to_bar and crowded:
        pt[idx] = B
      else:
        pt[idx] = S
    super().__init__(pt, seed)



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example that trains two agents using either LOLA or LOLA-DiCE.

An example that trains using LOLA (Foerster et al., 2017) or LOLA-DiCE
(Foerster et al., 2018) on iterated matrix games. Hyperparameters are
taken from the paper and https://github.com/alexis-jacq/LOLA_DiCE.
"""
import itertools
import os
import typing
from typing import List
from typing import Tuple
import warnings

from absl import app
from absl import flags
import distrax
import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np
import wandb

from open_spiel.python.environments.iterated_matrix_game import IteratedMatchingPennies
from open_spiel.python.environments.iterated_matrix_game import IteratedPrisonersDilemma
from open_spiel.python.jax.opponent_shaping import OpponentShapingAgent
from open_spiel.python.rl_environment import Environment
from open_spiel.python.rl_environment import TimeStep

warnings.simplefilter('ignore', FutureWarning)

FLAGS = flags.FLAGS
flags.DEFINE_string('exp_name', 'dice_1step_pytorchparams', 'Experiment name.')
flags.DEFINE_integer('seed', 42, 'Random seed.')
flags.DEFINE_string('game', 'ipd', 'Name of the game.')
flags.DEFINE_integer('epochs', 200, 'Number of training iterations.')
flags.DEFINE_integer('batch_size', 1024, 'Number of episodes in a batch.')
flags.DEFINE_integer(
    'critic_mini_batches', 1, 'Number of minibatches for critic.'
)
flags.DEFINE_integer('game_iterations', 150, 'Number of iterated plays.')
flags.DEFINE_float('policy_lr', 0.2, 'Policy learning rate.')
flags.DEFINE_float('opp_policy_lr', 0.3, 'Policy learning rate.')
flags.DEFINE_float('critic_lr', 0.1, 'Critic learning rate.')
flags.DEFINE_string('correction_type', 'lola', 'Either "lola", "dice" or None.')
flags.DEFINE_integer(
    'n_lookaheads', 2, 'Number of lookaheads for LOLA correction.'
)
flags.DEFINE_float(
    'correction_max_grad_norm',
    None,
    'Maximum gradient norm of LOLA correction.',
)
flags.DEFINE_float('discount', 0.96, 'Discount factor.')
flags.DEFINE_integer(
    'policy_update_interval',
    1,
    'Number of critic updates per before policy is updated.',
)
flags.DEFINE_integer('eval_batch_size', 1024, 'Random seed.')
flags.DEFINE_bool(
    'use_jit', False, 'If true, JAX jit compilation will be enabled.'
)
flags.DEFINE_bool(
    'use_opponent_modelling',
    True,
    'If false, ground truth opponent weights are used.',
)
flags.DEFINE_integer(
    'opp_policy_mini_batches', 8, 'Number of minibatches for opponent policy.'
)
flags.DEFINE_float(
    'opponent_model_learning_rate', 0.3, 'Learning rate for opponent model.'
)
flags.DEFINE_bool('debug', False, 'If true, debug mode is enabled.')


def get_action_probs(
    agent: OpponentShapingAgent, game: str
) -> List[typing.Dict[str, typing.Any]]:
  """Returns the probability of cooperation and a string repr for each state.
  
  Args:
      agent: The agent.
      game: The name of the game.

  Returns:
      A list of dictionaries, each containing the probability of cooperation
      and a string representation
  """
  actions = ['C', 'D'] if game == 'ipd' else ['H', 'T']
  states = ['s0'] + [''.join(s) for s in itertools.product(actions, repeat=2)]
  params = agent.train_state.policy_params[agent.player_id]
  action_probs = []
  for i, state_str in enumerate(states):
    state = np.eye(len(states))[i]
    prob = agent.policy_network.apply(params, state).prob(0)
    action = actions[0]
    action_probs.append(
        {'prob': prob.item(), 'name': f'P({action}|{state_str})'}
    )
  return action_probs


def log_epoch_data(epoch: int, agents: List[OpponentShapingAgent], eval_batch):
  """Logs data to wandb and prints it to the console.

  Args:
      epoch: The current epoch.
      agents: A list of agents.
      eval_batch: A batch of episodes.
  """
  logs = {}
  for agent in agents:
    avg_step_reward = np.mean(
        [ts.rewards[agent.player_id] for ts in eval_batch]
    )
    probs = get_action_probs(agent, game=FLAGS.game)
    for info in probs:
      logs[f'agent_{agent.player_id}/{info["name"]}'] = info['prob']
    probs = ', '.join([f'{info["name"]}: {info["prob"]:.2f}' for info in probs])
    metrics = agent.metrics()
    logs.update({
        f'agent_{agent.player_id}/avg_step_reward': avg_step_reward,
        **{
            f'agent_{agent.player_id}/{k}': v.item() for k, v in metrics.items()
        },
    })
    print(
        f'[epoch {epoch}] Agent {agent.player_id}: {avg_step_reward:.2f} |'
        f' {probs}'
    )
  wandb.log(logs)


def collect_batch(
    env: Environment, agents: List[OpponentShapingAgent], eval_mode: bool
) -> List[TimeStep]:
  """Collects one episode.

  Args:
      env: The environment.
      agents: A list of opponent shaping agents.
      eval_mode: If true, the agents will be run in evaluation mode.

  Returns:
      A list of time steps.
  """
  episode = []
  time_step = env.reset()
  episode.append(time_step)
  while not time_step.last():
    actions = []
    for agent in agents:
      action, _ = agent.step(time_step, is_evaluation=eval_mode)
      if action is not None:
        action = action.squeeze()
      actions.append(action)
    time_step = env.step(np.stack(actions, axis=1))
    time_step.observations['actions'] = actions
    episode.append(time_step)

  for agent in agents:
    agent.step(time_step, is_evaluation=eval_mode)
  return episode


def make_agent(
    key: jax.random.PRNGKey,
    player_id: int,
    env: Environment,
    networks: Tuple[hk.Transformed, hk.Transformed],
) -> OpponentShapingAgent:
  """Creates an opponent shaping agent.

  Args:
      key: A random seed key.
      player_id: The id of the player.
      env: The environment.
      networks: A tuple of policy and critic networks transformed by
        hk.transform.

  Returns:
      An opponent shaping agent instance.
  """
  policy_network, critic_network = networks
  return OpponentShapingAgent(
      player_id=player_id,
      opponent_ids=[1 - player_id],
      seed=key,
      info_state_size=env.observation_spec()['info_state'][player_id],
      num_actions=env.action_spec()['num_actions'][player_id],
      policy=policy_network,
      critic=critic_network,
      batch_size=FLAGS.batch_size,
      num_critic_mini_batches=FLAGS.critic_mini_batches,
      pi_learning_rate=FLAGS.policy_lr,
      opp_policy_learning_rate=FLAGS.opp_policy_lr,
      num_opponent_updates=FLAGS.opp_policy_mini_batches,
      critic_learning_rate=FLAGS.critic_lr,
      opponent_model_learning_rate=FLAGS.opponent_model_learning_rate,
      policy_update_interval=FLAGS.policy_update_interval,
      discount=FLAGS.discount,
      critic_discount=0,  # Predict the imm. reward (for iterated matrix games)
      correction_type=FLAGS.correction_type,
      clip_grad_norm=FLAGS.correction_max_grad_norm,
      use_jit=FLAGS.use_jit,
      n_lookaheads=FLAGS.n_lookaheads,
      env=env,
  )


def make_agent_networks(
    num_states: int, num_actions: int
) -> Tuple[hk.Transformed, hk.Transformed]:
  """Creates action weights for each state-action pair and values for each state.

  Args:
      num_states: The number of distinct states.
      num_actions: The number of distinct actions.

  Returns:
      A tuple of policy and critic networks transformed by hk.transform.
  """

  def policy(obs):
    theta = hk.get_parameter(
        'theta',
        init=hk.initializers.Constant(0),
        shape=(num_states, num_actions),
    )
    logits = jnp.select(obs, theta)
    logits = jnp.nan_to_num(logits)
    return distrax.Categorical(logits=logits)

  def value_fn(obs):
    w = hk.get_parameter(
        'w', [num_states], init=jnp.zeros
    )  # @pylint: disable=invalid-name
    return w[jnp.argmax(obs, axis=-1)].reshape(*obs.shape[:-1], 1)

  return hk.without_apply_rng(hk.transform(policy)), hk.without_apply_rng(
      hk.transform(value_fn)
  )


def make_env(game: str, iterations: int, batch_size: int) -> Environment:
  """Creates an environment.

  The environment is either iterated prisoners dilemma or iterated matching
  pennies.
  
  Args:
      game: The game to play. Either 'ipd' or 'imp'.
      iterations: The number of iterations to play.
      batch_size: The batch size.

  Returns:
      An environment instance.
  """
  if game == 'ipd':
    env = IteratedPrisonersDilemma(iterations=iterations, batch_size=batch_size)
  elif game == 'imp':
    env = IteratedMatchingPennies(iterations=iterations, batch_size=batch_size)
  else:
    raise ValueError(f'Unknown game: {game}')
  return env


def setup_agents(
    env: Environment, rng: hk.PRNGSequence
) -> List[OpponentShapingAgent]:
  """Creates an opponent shaping agent for each player in the environment.

  Args:
      env: The environment.
      rng: A random seed key.

  Returns:
      A list of opponent shaping agents.
  """
  agents = []
  num_actions = env.action_spec()['num_actions']
  info_state_shape = env.observation_spec()['info_state']
  for player_id in range(env.num_players):
    networks = make_agent_networks(
        num_states=info_state_shape[player_id][0],
        num_actions=num_actions[player_id],
    )
    agent = make_agent(
        key=next(rng), player_id=player_id, env=env, networks=networks
    )
    agents.append(agent)
  return agents


def update_weights(agents: List[OpponentShapingAgent]):
  """Updates the weights of the opponent models.

  Args:
      agents: A list of opponent shaping agents.

  Returns:
      None
  """
  agent: OpponentShapingAgent
  for agent in agents:
    for opp in [a for a in agents if a.player_id != agent.player_id]:
      agent.update_params(state=opp.train_state, player_id=opp.player_id)


def main(_):
  """Main function. Runs the experiment."""
  if FLAGS.exp_name is None:
    FLAGS.exp_name = f'{FLAGS.game}_{FLAGS.seed}'
  if not FLAGS.debug:
    wandb.login(key=os.environ.get('WANDB_API_KEY', None))
  wandb.init(
      project='open-spiel-opponent-modelling',
      group=FLAGS.exp_name,
      config={
          'game': FLAGS.game,
          'seed': FLAGS.seed,
          'epochs': FLAGS.epochs,
          'batch_size': FLAGS.batch_size,
          'critic_mini_batches': FLAGS.critic_mini_batches,
          'game_iterations': FLAGS.game_iterations,
          'policy_lr': FLAGS.policy_lr,
          'opp_policy_lr': FLAGS.opp_policy_lr,
          'critic_lr': FLAGS.critic_lr,
          'correction_type': FLAGS.correction_type,
          'n_lookaheads': FLAGS.n_lookaheads,
          'correction_max_grad_norm': FLAGS.correction_max_grad_norm,
          'discount': FLAGS.discount,
          'policy_update_interval': FLAGS.policy_update_interval,
          'use_opponent_modelling': FLAGS.use_opponent_modelling,
          'opp_policy_mini_batches': FLAGS.opp_policy_mini_batches,
          'opponent_model_learning_rate': FLAGS.opponent_model_learning_rate,
      },
      mode='disabled' if FLAGS.debug else 'online',
  )

  rng = hk.PRNGSequence(key_or_seed=FLAGS.seed)
  env = make_env(
      iterations=FLAGS.game_iterations,
      batch_size=FLAGS.batch_size,
      game=FLAGS.game,
  )
  agents = setup_agents(env=env, rng=rng)

  if not FLAGS.use_opponent_modelling:
    update_weights(agents)

  batch = collect_batch(env=env, agents=agents, eval_mode=True)
  log_epoch_data(epoch=0, agents=agents, eval_batch=batch)
  for epoch in range(1, FLAGS.epochs + 1):
    batch = collect_batch(env=env, agents=agents, eval_mode=False)
    if not FLAGS.use_opponent_modelling:
      update_weights(agents)
    log_epoch_data(epoch=epoch, agents=agents, eval_batch=batch)
    print('#' * 100)

  wandb.finish()


if __name__ == '__main__':
  app.run(main)



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A tabular representation of a distribution for a game."""

from typing import Dict, Optional

from open_spiel.python.mfg import distribution
import pyspiel

DistributionDict = Dict[str, float]


class TabularDistribution(distribution.ParametricDistribution):
  """Distribution that uses a dictionary to store the values of the states."""

  def __init__(self, game: pyspiel.Game):
    self._distribution: DistributionDict = {}
    super().__init__(game)

  def value(self, state: pyspiel.State) -> float:
    return self.value_str(self.state_to_str(state))

  def value_str(self,
                state_str: str,
                default_value: Optional[float] = None) -> float:
    """Returns the probability of the distribution on the state string given.

    Args:
      state_str: A string.
      default_value: If not None, return this value if the state is not in the
        support of the distribution.

    Returns:
      A `float`.

    Raises:
      ValueError: If the state has not been seen by the distribution and no
        default value has been passed to the method.
    """
    if default_value is None:
      try:
        return self._distribution[state_str]
      except KeyError as e:
        raise ValueError(
            f"Distribution not computed for state {state_str}") from e
    return self._distribution.get(state_str, default_value)

  def get_params(self) -> DistributionDict:
    return self._distribution

  def set_params(self, params: DistributionDict):
    self._distribution = params

  def state_to_str(self, state: pyspiel.State) -> str:
    # TODO(author15): Consider switching to
    # state.mean_field_population(). For now, this does not matter in
    # practice since games don't have different observation strings for
    # different player IDs.
    return state.observation_string(pyspiel.PlayerId.DEFAULT_PLAYER_ID)

  @property
  def distribution(self) -> DistributionDict:
    return self._distribution



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example use of the CFR algorithm on Kuhn Poker."""

from absl import app

from open_spiel.python.algorithms import cfr
from open_spiel.python.algorithms import expected_game_score
import pyspiel


def main(_):
  game = pyspiel.load_game("kuhn_poker")

  cfr_solver = cfr.CFRSolver(game)
  iterations = 1000

  for i in range(iterations):
    cfr_value = cfr_solver.evaluate_and_update_policy()
    print("Game util at iteration {}: {}".format(i, cfr_value))

  average_policy = cfr_solver.average_policy()
  average_policy_values = expected_game_score.policy_value(
      game.new_initial_state(), [average_policy] * 2)
  print("Computed player 0 value: {}".format(average_policy_values[0]))
  print("Expected player 0 value: {}".format(-1 / 18))


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Python implementation of the counterfactual regret minimization algorithm.

One iteration of CFR consists of:
1) Compute current strategy from regrets (e.g. using Regret Matching).
2) Compute values using the current strategy
3) Compute regrets from these values

The average policy is what converges to a Nash Equilibrium.
"""

import collections
import attr
import numpy as np

from open_spiel.python import policy
import pyspiel


@attr.s
class _InfoStateNode(object):
  """An object wrapping values associated to an information state."""
  # The list of the legal actions.
  legal_actions = attr.ib()
  index_in_tabular_policy = attr.ib()
  # Map from information states string representations and actions to the
  # counterfactual regrets, accumulated over the policy iterations
  cumulative_regret = attr.ib(factory=lambda: collections.defaultdict(float))
  # Same as above for the cumulative of the policy probabilities computed
  # during the policy iterations
  cumulative_policy = attr.ib(factory=lambda: collections.defaultdict(float))


def _apply_regret_matching_plus_reset(info_state_nodes):
  """Resets negative cumulative regrets to 0.

  Regret Matching+ corresponds to the following cumulative regrets update:
  cumulative_regrets = max(cumulative_regrets + regrets, 0)

  This must be done at the level of the information set, and thus cannot be
  done during the tree traversal (which is done on histories). It is thus
  performed as an additional step.

  This function is a module level function to be reused by both CFRSolver and
  CFRBRSolver.

  Args:
    info_state_nodes: A dictionary {`info_state_str` -> `_InfoStateNode`}.
  """
  for info_state_node in info_state_nodes.values():
    action_to_cum_regret = info_state_node.cumulative_regret
    for action, cumulative_regret in action_to_cum_regret.items():
      if cumulative_regret < 0:
        action_to_cum_regret[action] = 0


def _update_current_policy(current_policy, info_state_nodes):
  """Updates in place `current_policy` from the cumulative regrets.

  This function is a module level function to be reused by both CFRSolver and
  CFRBRSolver.

  Args:
    current_policy: A `policy.TabularPolicy` to be updated in-place.
    info_state_nodes: A dictionary {`info_state_str` -> `_InfoStateNode`}.
  """
  for info_state, info_state_node in info_state_nodes.items():
    state_policy = current_policy.policy_for_key(info_state)

    for action, value in _regret_matching(
        info_state_node.cumulative_regret,
        info_state_node.legal_actions).items():
      state_policy[action] = value


def _update_average_policy(average_policy, info_state_nodes):
  """Updates in place `average_policy` to the average of all policies iterated.

  This function is a module level function to be reused by both CFRSolver and
  CFRBRSolver.

  Args:
    average_policy: A `policy.TabularPolicy` to be updated in-place.
    info_state_nodes: A dictionary {`info_state_str` -> `_InfoStateNode`}.
  """
  for info_state, info_state_node in info_state_nodes.items():
    info_state_policies_sum = info_state_node.cumulative_policy
    state_policy = average_policy.policy_for_key(info_state)
    probabilities_sum = sum(info_state_policies_sum.values())
    if probabilities_sum == 0:
      num_actions = len(info_state_node.legal_actions)
      for action in info_state_node.legal_actions:
        state_policy[action] = 1 / num_actions
    else:
      for action, action_prob_sum in info_state_policies_sum.items():
        state_policy[action] = action_prob_sum / probabilities_sum


class _CFRSolverBase(object):
  r"""A base class for both CFR and CFR-BR.

  The main iteration loop is implemented in `evaluate_and_update_policy`:

  ```python
      game = pyspiel.load_game("game_name")
      initial_state = game.new_initial_state()

      solver = Solver(game)

      for i in range(num_iterations):
        solver.evaluate_and_update_policy()
        solver.current_policy()  # Access the current policy
        solver.average_policy()  # Access the average policy
  ```
  """

  def __init__(self, game, alternating_updates, linear_averaging,
               regret_matching_plus):
    # pyformat: disable
    """Initializer.

    Args:
      game: The `pyspiel.Game` to run on.
      alternating_updates: If `True`, alternating updates are performed: for
        each player, we compute and update the cumulative regrets and policies.
        In that case, and when the policy is frozen during tree traversal, the
        cache is reset after each update for one player.
        Otherwise, the update is simultaneous.
      linear_averaging: Whether to use linear averaging, i.e.
        cumulative_policy[info_state][action] += (
          iteration_number * reach_prob * action_prob)

        or not:

        cumulative_policy[info_state][action] += reach_prob * action_prob
      regret_matching_plus: Whether to use Regret Matching+:
        cumulative_regrets = max(cumulative_regrets + regrets, 0)
        or simply regret matching:
        cumulative_regrets = cumulative_regrets + regrets
    """
    # pyformat: enable
    assert game.get_type().dynamics == pyspiel.GameType.Dynamics.SEQUENTIAL, (
        "CFR requires sequential games. If you're trying to run it " +
        "on a simultaneous (or normal-form) game, please first transform it " +
        "using turn_based_simultaneous_game.")

    self._game = game
    self._num_players = game.num_players()
    self._root_node = self._game.new_initial_state()

    # This is for returning the current policy and average policy to a caller
    self._current_policy = policy.TabularPolicy(game)
    self._average_policy = self._current_policy.__copy__()

    self._info_state_nodes = {}
    self._initialize_info_state_nodes(self._root_node)

    self._iteration = 0  # For possible linear-averaging.
    self._linear_averaging = linear_averaging
    self._alternating_updates = alternating_updates
    self._regret_matching_plus = regret_matching_plus

  def _initialize_info_state_nodes(self, state):
    """Initializes info_state_nodes.

    Create one _InfoStateNode per infoset. We could also initialize the node
    when we try to access it and it does not exist.

    Args:
      state: The current state in the tree walk. This should be the root node
        when we call this function from a CFR solver.
    """
    if state.is_terminal():
      return

    if state.is_chance_node():
      for action, unused_action_prob in state.chance_outcomes():
        self._initialize_info_state_nodes(state.child(action))
      return

    current_player = state.current_player()
    info_state = state.information_state_string(current_player)

    info_state_node = self._info_state_nodes.get(info_state)
    if info_state_node is None:
      legal_actions = state.legal_actions(current_player)
      info_state_node = _InfoStateNode(
          legal_actions=legal_actions,
          index_in_tabular_policy=self._current_policy.state_lookup[info_state])
      self._info_state_nodes[info_state] = info_state_node

    for action in info_state_node.legal_actions:
      self._initialize_info_state_nodes(state.child(action))

  def current_policy(self):
    """Returns the current policy as a TabularPolicy.

    WARNING: The same object, updated in-place will be returned! You can copy
    it (or its `action_probability_array` field).

    For CFR/CFR+, this policy does not necessarily have to converge. It
    converges with high probability for CFR-BR.
    """
    return self._current_policy

  def average_policy(self):
    """Returns the average of all policies iterated.

    WARNING: The same object, updated in-place will be returned! You can copy
    it (or its `action_probability_array` field).

    This average policy converges to a Nash policy as the number of iterations
    increases.

    The policy is computed using the accumulated policy probabilities computed
    using `evaluate_and_update_policy`.

    Returns:
      A `policy.TabularPolicy` object (shared between calls) giving the (linear)
      time averaged policy (weighted by player reach probabilities) for both
      players.
    """
    _update_average_policy(self._average_policy, self._info_state_nodes)
    return self._average_policy

  def _compute_counterfactual_regret_for_player(self, state, policies,
                                                reach_probabilities, player):
    """Increments the cumulative regrets and policy for `player`.

    Args:
      state: The initial game state to analyze from.
      policies: A list of `num_players` callables taking as input an
        `info_state_node` and returning a {action: prob} dictionary. For CFR,
          this is simply returning the current policy, but this can be used in
          the CFR-BR solver, to prevent code duplication. If None,
          `_get_infostate_policy` is used.
      reach_probabilities: The probability for each player of reaching `state`
        as a numpy array [prob for player 0, for player 1,..., for chance].
        `player_reach_probabilities[player]` will work in all cases.
      player: The 0-indexed player to update the values for. If `None`, the
        update for all players will be performed.

    Returns:
      The utility of `state` for all players, assuming all players follow the
      current policy defined by `self.Policy`.
    """
    if state.is_terminal():
      return np.asarray(state.returns())

    if state.is_chance_node():
      state_value = 0.0
      for action, action_prob in state.chance_outcomes():
        assert action_prob > 0
        new_state = state.child(action)
        new_reach_probabilities = reach_probabilities.copy()
        new_reach_probabilities[-1] *= action_prob
        state_value += action_prob * self._compute_counterfactual_regret_for_player(
            new_state, policies, new_reach_probabilities, player)
      return state_value

    current_player = state.current_player()
    info_state = state.information_state_string(current_player)

    # No need to continue on this history branch as no update will be performed
    # for any player.
    # The value we return here is not used in practice. If the conditional
    # statement is True, then the last taken action has probability 0 of
    # occurring, so the returned value is not impacting the parent node value.
    if all(reach_probabilities[:-1] == 0):
      return np.zeros(self._num_players)

    state_value = np.zeros(self._num_players)

    # The utilities of the children states are computed recursively. As the
    # regrets are added to the information state regrets for each state in that
    # information state, the recursive call can only be made once per child
    # state. Therefore, the utilities are cached.
    children_utilities = {}

    info_state_node = self._info_state_nodes[info_state]
    if policies is None:
      info_state_policy = self._get_infostate_policy(info_state)
    else:
      info_state_policy = policies[current_player](info_state)
    for action in state.legal_actions():
      action_prob = info_state_policy.get(action, 0.)
      new_state = state.child(action)
      new_reach_probabilities = reach_probabilities.copy()
      new_reach_probabilities[current_player] *= action_prob
      child_utility = self._compute_counterfactual_regret_for_player(
          new_state,
          policies=policies,
          reach_probabilities=new_reach_probabilities,
          player=player)

      state_value += action_prob * child_utility
      children_utilities[action] = child_utility

    # If we are performing alternating updates, and the current player is not
    # the current_player, we skip the cumulative values update.
    # If we are performing simultaneous updates, we do update the cumulative
    # values.
    simulatenous_updates = player is None
    if not simulatenous_updates and current_player != player:
      return state_value

    reach_prob = reach_probabilities[current_player]
    counterfactual_reach_prob = (
        np.prod(reach_probabilities[:current_player]) *
        np.prod(reach_probabilities[current_player + 1:]))
    state_value_for_player = state_value[current_player]

    for action, action_prob in info_state_policy.items():
      cfr_regret = counterfactual_reach_prob * (
          children_utilities[action][current_player] - state_value_for_player)

      info_state_node.cumulative_regret[action] += cfr_regret
      if self._linear_averaging:
        info_state_node.cumulative_policy[
            action] += self._iteration * reach_prob * action_prob
      else:
        info_state_node.cumulative_policy[action] += reach_prob * action_prob

    return state_value

  def _get_infostate_policy(self, info_state_str):
    """Returns an {action: prob} dictionary for the policy on `info_state`."""
    info_state_node = self._info_state_nodes[info_state_str]
    prob_vec = self._current_policy.action_probability_array[
        info_state_node.index_in_tabular_policy]
    return {
        action: prob_vec[action] for action in info_state_node.legal_actions
    }


def _regret_matching(cumulative_regrets, legal_actions):
  """Returns an info state policy by applying regret-matching.

  Args:
    cumulative_regrets: A {action: cumulative_regret} dictionary.
    legal_actions: the list of legal actions at this state.

  Returns:
    A dict of action -> prob for all legal actions.
  """
  regrets = cumulative_regrets.values()
  sum_positive_regrets = sum((regret for regret in regrets if regret > 0))

  info_state_policy = {}
  if sum_positive_regrets > 0:
    for action in legal_actions:
      positive_action_regret = max(0.0, cumulative_regrets[action])
      info_state_policy[action] = (
          positive_action_regret / sum_positive_regrets)
  else:
    for action in legal_actions:
      info_state_policy[action] = 1.0 / len(legal_actions)
  return info_state_policy


class _CFRSolver(_CFRSolverBase):
  r"""Implements the Counterfactual Regret Minimization (CFR) algorithm.

  The algorithm computes an approximate Nash policy for 2 player zero-sum games.

  CFR can be view as a policy iteration algorithm. Importantly, the policies
  themselves do not converge to a Nash policy, but their average does.

  The main iteration loop is implemented in `evaluate_and_update_policy`:

  ```python
      game = pyspiel.load_game("game_name")
      initial_state = game.new_initial_state()

      cfr_solver = CFRSolver(game)

      for i in range(num_iterations):
        cfr.evaluate_and_update_policy()
  ```

  Once the policy has converged, the average policy (which converges to the Nash
  policy) can be computed:
  ```python
        average_policy = cfr_solver.ComputeAveragePolicy()
  ```

  # Policy and average policy

  policy(0) and average_policy(0) are not technically defined, but these
  methods will return arbitrarily the uniform_policy.

  Then, we are expected to have:

  ```
  for t in range(1, N):
    cfr_solver.evaluate_and_update_policy()
    policy(t) = RM or RM+ of cumulative regrets
    avg_policy(t)(s, a) ~ \sum_{k=1}^t player_reach_prob(t)(s) * policy(k)(s, a)

    With Linear Averaging, the avg_policy is proportional to:
    \sum_{k=1}^t k * player_reach_prob(t)(s) * policy(k)(s, a)
  ```
  """

  def evaluate_and_update_policy(self):
    """Performs a single step of policy evaluation and policy improvement."""
    self._iteration += 1
    if self._alternating_updates:
      for player in range(self._game.num_players()):
        self._compute_counterfactual_regret_for_player(
            self._root_node,
            policies=None,
            reach_probabilities=np.ones(self._game.num_players() + 1),
            player=player)
        if self._regret_matching_plus:
          _apply_regret_matching_plus_reset(self._info_state_nodes)
        _update_current_policy(self._current_policy, self._info_state_nodes)
    else:
      self._compute_counterfactual_regret_for_player(
          self._root_node,
          policies=None,
          reach_probabilities=np.ones(self._game.num_players() + 1),
          player=None)
      if self._regret_matching_plus:
        _apply_regret_matching_plus_reset(self._info_state_nodes)
      _update_current_policy(self._current_policy, self._info_state_nodes)


class CFRPlusSolver(_CFRSolver):
  """CFR+ implementation.

  The algorithm computes an approximate Nash policy for 2 player zero-sum games.
  More generally, it should approach a no-regret set, which corresponds to the
  set of coarse-correlated equilibria. See https://arxiv.org/abs/1305.0034

  CFR can be view as a policy iteration algorithm. Importantly, the policies
  themselves do not converge to a Nash policy, but their average does.

  See https://poker.cs.ualberta.ca/publications/2015-ijcai-cfrplus.pdf

  CFR+ is CFR with the following modifications:
  - use Regret Matching+ instead of Regret Matching.
  - use alternating updates instead of simultaneous updates.
  - use linear averaging.

  Usage:

  ```python
      game = pyspiel.load_game("game_name")
      initial_state = game.new_initial_state()

      cfr_solver = CFRSolver(game)

      for i in range(num_iterations):
        cfr.evaluate_and_update_policy()
  ```

  Once the policy has converged, the average policy (which converges to the Nash
  policy) can be computed:
  ```python
        average_policy = cfr_solver.ComputeAveragePolicy()
  ```
  """

  def __init__(self, game):
    super(CFRPlusSolver, self).__init__(
        game,
        regret_matching_plus=True,
        alternating_updates=True,
        linear_averaging=True)


class CFRSolver(_CFRSolver):
  """Implements the Counterfactual Regret Minimization (CFR) algorithm.

  See https://poker.cs.ualberta.ca/publications/NIPS07-cfr.pdf

  NOTE: We use alternating updates (which was not the case in the original
  paper) because it has been proved to be far more efficient.
  """

  def __init__(self, game):
    super(CFRSolver, self).__init__(
        game,
        regret_matching_plus=False,
        alternating_updates=True,
        linear_averaging=False)



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""RNN model."""

from typing import Callable, List, Union, Optional

import haiku as hk
import jax.numpy as jnp


class RNNModel(hk.RNNCore):
  """RNN model."""

  def __init__(self,
               layers: List[Union[hk.Module, Callable[[jnp.ndarray],
                                                      jnp.ndarray]]],
               name: Optional[str] = 'RNN'):
    super().__init__(name=name)
    self._layers = layers

  def __call__(self, inputs, prev_state):
    x = inputs
    curr_state = [None] * len(prev_state)
    for k, layer in enumerate(self._layers):
      if isinstance(layer, hk.RNNCore):
        x, curr_state[k] = layer(x, prev_state[k])
      else:
        x = layer(x)
    return x, tuple(curr_state)

  def initial_state(self, batch_size: Optional[int]):
    layerwise_init_state = []
    for layer in self._layers:
      if isinstance(layer, hk.RNNCore):
        layerwise_init_state.append(layer.initial_state(batch_size))
      else:
        layerwise_init_state.append(None)
    return tuple(layerwise_init_state)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.algorithms.psro_v2.strategy_selectors."""

from absl.testing import absltest
import numpy as np
from open_spiel.python.algorithms.psro_v2 import strategy_selectors


class FakeSolver(object):

  def __init__(self, strategies, policies):
    self.strategies = strategies
    self.policies = policies

  def get_policies(self):
    return self.policies

  def get_meta_strategies(self):
    return self.strategies


def equal_to_transposition_lists(a, b):
  return [set(x) for x in a] == [set(x) for x in b]


EPSILON_MIN_POSITIVE_PROBA = 1e-8


def rectified_alias(solver, number_policies_to_select):
  """Returns every strategy with nonzero selection probability.

  Args:
    solver: A GenPSROSolver instance.
    number_policies_to_select: Number policies to select

  Returns:
    used_policies: A list, each element a list of the policies used per player.
  """
  del number_policies_to_select

  used_policies = []
  used_policy_indexes = []

  policies = solver.get_policies()
  num_players = len(policies)
  meta_strategy_probabilities = solver.get_meta_strategies()

  for k in range(num_players):
    current_policies = policies[k]
    current_probabilities = meta_strategy_probabilities[k]

    current_indexes = [
        i for i in range(len(current_policies))
        if current_probabilities[i] > EPSILON_MIN_POSITIVE_PROBA
    ]
    current_policies = [
        current_policies[i]
        for i in current_indexes
    ]

    used_policy_indexes.append(current_indexes)
    used_policies.append(current_policies)
  return used_policies, used_policy_indexes


def probabilistic_alias(solver, number_policies_to_select):
  """Returns [kwargs] policies randomly, proportionally with selection probas.

  Args:
    solver: A GenPSROSolver instance.
    number_policies_to_select: Number policies to select
  """
  policies = solver.get_policies()
  num_players = len(policies)
  meta_strategy_probabilities = solver.get_meta_strategies()

  print(policies, meta_strategy_probabilities)
  used_policies = []
  used_policy_indexes = []
  for k in range(num_players):
    current_policies = policies[k]
    current_selection_probabilities = meta_strategy_probabilities[k]
    effective_number = min(number_policies_to_select, len(current_policies))

    selected_indexes = list(
        np.random.choice(
            list(range(len(current_policies))),
            effective_number,
            replace=False,
            p=current_selection_probabilities))
    selected_policies = [current_policies[i] for i in selected_indexes]
    used_policies.append(selected_policies)
    used_policy_indexes.append(selected_indexes)
  return used_policies, used_policy_indexes


def top_k_probabilities_alias(solver, number_policies_to_select):
  """Returns [kwargs] policies with highest selection probabilities.

  Args:
    solver: A GenPSROSolver instance.
    number_policies_to_select: Number policies to select
  """
  policies = solver.get_policies()
  num_players = len(policies)
  meta_strategy_probabilities = solver.get_meta_strategies()

  used_policies = []
  used_policy_indexes = []
  for k in range(num_players):
    current_policies = policies[k]
    current_selection_probabilities = meta_strategy_probabilities[k]
    effective_number = min(number_policies_to_select, len(current_policies))

    # pylint: disable=g-complex-comprehension
    selected_indexes = [
        index for _, index in sorted(
            zip(current_selection_probabilities,
                list(range(len(current_policies)))),
            key=lambda pair: pair[0])
    ][:effective_number]

    selected_policies = [current_policies[i] for i in selected_indexes]
    used_policies.append(selected_policies)
    used_policy_indexes.append(selected_indexes)
  return used_policies, used_policy_indexes


class StrategySelectorsTest(absltest.TestCase):

  def test_vital(self):
    n_tests = 1000
    number_strategies = 50
    number_players = 3
    for i in range(n_tests):
      probabilities = np.random.uniform(size=(number_players,
                                              number_strategies))
      probabilities /= np.sum(probabilities, axis=1).reshape(-1, 1)
      probabilities = list(probabilities)

      policies = [list(range(number_strategies)) for _ in range(number_players)]

      solver = FakeSolver(probabilities, policies)

      # To see how rectified reacts to 0 probabilities.
      probabilities[0][0] = 0
      probabilities[-1][-1] = 0
      a, b = strategy_selectors.rectified(solver, 1)
      c, d = rectified_alias(solver, 1)

      self.assertEqual(a, c, "Rectified failed.")
      self.assertEqual(b, d, "Rectified failed.")

      a, b = strategy_selectors.top_k_probabilities(solver, 3)
      c, d = top_k_probabilities_alias(solver, 3)

      self.assertEqual(a, c, "Top k failed.")
      self.assertEqual(b, d, "Top k failed.")

      n_nonzero_policies = 2
      probabilities = [np.zeros(number_strategies) for _ in range(
          number_players)]

      for player in range(number_players):
        for _ in range(n_nonzero_policies):
          i = np.random.randint(0, high=number_strategies)
          while probabilities[player][i] > 1e-12:
            i = np.random.randint(0, high=number_strategies)
          probabilities[player][i] = 1.0 / n_nonzero_policies
        probabilities[player] /= np.sum(probabilities[player])

      solver = FakeSolver(probabilities, policies)
      a, b = strategy_selectors.probabilistic(solver, n_nonzero_policies)
      c, d = probabilistic_alias(solver, n_nonzero_policies)

      self.assertTrue(equal_to_transposition_lists(a, c),
                      "Probabilistic failed.")
      self.assertTrue(equal_to_transposition_lists(b, d),
                      "Probabilistic failed.")

if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.algorithms.losses.rl_losses."""

from absl.testing import parameterized
import numpy as np
import tensorflow.compat.v1 as tf

from open_spiel.python.algorithms.losses import rl_losses

# Temporarily disable v2 behavior until code is updated.
tf.disable_v2_behavior()


class RLLossesTest(parameterized.TestCase, tf.test.TestCase):

  @parameterized.named_parameters(('no_entropy_cost', 0.),
                                  ('with_entropy_cost', 1.))
  def test_batch_qpg_loss_with_entropy_cost(self, entropy_cost):
    batch_qpg_loss = rl_losses.BatchQPGLoss(entropy_cost=entropy_cost)
    q_values = tf.constant([[0., -1., 1.], [1., -1., 0]], dtype=tf.float32)
    policy_logits = tf.constant([[1., 1., 1.], [1., 1., 4.]], dtype=tf.float32)
    total_loss = batch_qpg_loss.loss(policy_logits, q_values)
    # Compute expected quantities.
    expected_policy_entropy_loss = -1 * (1.0986 + 0.3665) / 2
    # baseline = \sum_a pi_a * Q_a = 0.
    # -\sum_a pi_a * (Q_a - baseline)
    expected_policy_loss = (0.0 + 0.0) / 2
    expected_total_loss = (
        expected_policy_loss + entropy_cost * expected_policy_entropy_loss)
    with self.session() as sess:
      np.testing.assert_allclose(
          sess.run(total_loss), expected_total_loss, atol=1e-4)

  @parameterized.named_parameters(('no_entropy_cost', 0.),
                                  ('with_entropy_cost', 1.))
  def test_batch_rm_loss_with_entropy_cost(self, entropy_cost):
    batch_rpg_loss = rl_losses.BatchRMLoss(entropy_cost=entropy_cost)
    q_values = tf.constant([[0., -1., 1.], [1., -1., 0]], dtype=tf.float32)
    policy_logits = tf.constant([[1., 1., 1.], [1., 1., 4.]], dtype=tf.float32)
    total_loss = batch_rpg_loss.loss(policy_logits, q_values)
    # Compute expected quantities.
    expected_policy_entropy_loss = -(1.0986 + 0.3665) / 2
    # baseline = \sum_a pi_a * Q_a = 0.
    # -\sum_a pi_a * relu(Q_a - baseline)
    # negative sign as it's a loss term and loss needs to be minimized.
    expected_policy_loss = -(.3333 + .0452) / 2
    expected_total_loss = (
        expected_policy_loss + entropy_cost * expected_policy_entropy_loss)
    with self.session() as sess:
      np.testing.assert_allclose(
          sess.run(total_loss), expected_total_loss, atol=1e-3)

  @parameterized.named_parameters(('no_entropy_cost', 0.),
                                  ('with_entropy_cost', 1.))
  def test_batch_rpg_loss_with_entropy_cost(self, entropy_cost):
    batch_rpg_loss = rl_losses.BatchRPGLoss(entropy_cost=entropy_cost)
    q_values = tf.constant([[0., -1., 1.], [1., -1., 0]], dtype=tf.float32)
    policy_logits = tf.constant([[1., 1., 1.], [1., 1., 4.]], dtype=tf.float32)
    total_loss = batch_rpg_loss.loss(policy_logits, q_values)
    # Compute expected quantities.
    expected_policy_entropy_loss = -1 * (1.0986 + 0.3665) / 2
    # baseline = \sum_a pi_a * Q_a = 0.
    # \sum_a relu(Q_a - baseline)
    expected_policy_loss = (1.0 + 1.0) / 2
    expected_total_loss = (
        expected_policy_loss + entropy_cost * expected_policy_entropy_loss)
    with self.session() as sess:
      np.testing.assert_allclose(
          sess.run(total_loss), expected_total_loss, atol=1e-4)

  @parameterized.named_parameters(('no_entropy_cost', 0.),
                                  ('with_entropy_cost', 1.))
  def test_batch_a2c_loss_with_entropy_cost(self, entropy_cost):
    batch_a2c_loss = rl_losses.BatchA2CLoss(entropy_cost=entropy_cost)
    policy_logits = tf.constant([[1., 1., 1.], [1., 1., 4.]], dtype=tf.float32)
    baseline = tf.constant([1. / 3, 0.5], dtype=tf.float32)
    actions = tf.constant([1, 2], dtype=tf.int32)
    returns = tf.constant([0., 1.], dtype=tf.float32)
    total_loss = batch_a2c_loss.loss(policy_logits, baseline, actions, returns)
    # Compute expected quantities.
    # advantages = returns - baseline = [-1./3, 0.5]
    # cross_entropy = [-log(e^1./3 * e^1), -log(e^4/(e^4+ e + e))]
    # = [1.0986, 0.09492]
    # policy_loss = cross_entropy * advantages = [-0.3662, 0.04746]
    expected_policy_entropy_loss = -1 * (1.0986 + 0.3665) / 2
    expected_policy_loss = (-0.3662 + 0.04746) / 2
    expected_total_loss = (
        expected_policy_loss + entropy_cost * expected_policy_entropy_loss)
    with self.session() as sess:
      np.testing.assert_allclose(
          sess.run(total_loss), expected_total_loss, atol=1e-4)


if __name__ == '__main__':
  tf.test.main()



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for Fixed Point."""

from absl.testing import absltest
from absl.testing import parameterized

from open_spiel.python.mfg.algorithms import fixed_point
from open_spiel.python.mfg.algorithms import nash_conv
from open_spiel.python.mfg.games import crowd_modelling  # pylint: disable=unused-import
import pyspiel


class FixedPointTest(parameterized.TestCase):

  @parameterized.named_parameters(('python', 'python_mfg_crowd_modelling'),
                                  ('cpp', 'mfg_crowd_modelling'))
  def test_run(self, name):
    """Checks if the algorithm works."""
    game = pyspiel.load_game(name)
    fixed_p = fixed_point.FixedPoint(game)

    for _ in range(10):
      fixed_p.iteration()

    fixed_p_policy = fixed_p.get_policy()
    nash_conv_fixed_p = nash_conv.NashConv(game, fixed_p_policy)

    self.assertAlmostEqual(nash_conv_fixed_p.nash_conv(), 55.745, places=3)

  @parameterized.named_parameters(('python', 'python_mfg_crowd_modelling'),
                                  ('cpp', 'mfg_crowd_modelling'))
  def test_softmax(self, name):
    """Checks the softmax policy."""
    game = pyspiel.load_game(name)
    fixed_p = fixed_point.FixedPoint(game, temperature=10.0)

    for _ in range(10):
      fixed_p.iteration()

    fixed_p_policy = fixed_p.get_policy()
    nash_conv_fixed_p = nash_conv.NashConv(game, fixed_p_policy)

    self.assertAlmostEqual(nash_conv_fixed_p.nash_conv(), 2.421, places=3)


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.algorithms.adidas_utils.helpers.symmetric.utils."""

from absl import logging  # pylint:disable=unused-import
from absl.testing import absltest
from absl.testing import parameterized

import numpy as np

from open_spiel.python.algorithms.adidas_utils.helpers.symmetric import utils


class UtilsTest(parameterized.TestCase):

  def test_symmetrize_tensor(self, trials=100, seed=1234):
    random = np.random.RandomState(seed)

    successes = []
    for _ in range(trials):
      pt = random.rand(3, 2, 2, 2)

      pt_sym_man = np.zeros_like(pt)
      for p in range(3):
        for i in range(2):
          for j in range(2):
            for k in range(2):
              if p == 0:
                # read: if player 0 plays i and its two opponents play j and k
                # this should return the same payoff as when
                # player 1 plays i and its two opponents play j and k
                # player 2 plays i and its two opponents play j and k
                # solution is to add up all these payoffs and replace with avg
                pt_sym_man[p, i, j, k] = (pt[0, i, j, k] + pt[0, i, k, j] +
                                          pt[1, j, i, k] + pt[1, k, i, j] +
                                          pt[2, j, k, i] + pt[2, k, j, i]) / 6.
              elif p == 1:
                # same rationale, but with player 1 playing j
                pt_sym_man[p, i, j, k] = (pt[0, j, i, k] + pt[0, j, k, i] +
                                          pt[1, i, j, k] + pt[1, k, j, i] +
                                          pt[2, i, k, j] + pt[2, k, i, j]) / 6.
              else:
                # same rationale, but with player 2 playing k
                pt_sym_man[p, i, j, k] = (pt[0, k, i, j] + pt[0, k, j, i] +
                                          pt[1, i, k, j] + pt[1, j, k, i] +
                                          pt[2, i, j, k] + pt[2, j, i, k]) / 6.
      pt_sym = utils.sym(pt)

      successes += [np.allclose(pt_sym, pt_sym_man)]

    perc = 100 * np.mean(successes)
    logging.info('symmetrizing success rate out of %d is %f', trials, perc)
    self.assertGreaterEqual(
        perc, 100., 'symmetrizing failed')


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Liar's Poker implemented in Python."""

import numpy as np

import pyspiel


CHALLENGE_ACTION = 0
BID_ACTION_OFFSET = 1

_MAX_NUM_PLAYERS = 10
_MIN_NUM_PLAYERS = 2
_HAND_LENGTH = 10
_NUM_DIGITS = 10  # Number of digits to include from the range 1, 2, ..., 9, 0
_FULL_DECK = [1, 2, 3, 4, 5, 6, 7, 8, 9, 0]

_GAME_TYPE = pyspiel.GameType(
    short_name="python_liars_poker",
    long_name="Python Liars Poker",
    dynamics=pyspiel.GameType.Dynamics.SEQUENTIAL,
    chance_mode=pyspiel.GameType.ChanceMode.EXPLICIT_STOCHASTIC,
    information=pyspiel.GameType.Information.IMPERFECT_INFORMATION,
    utility=pyspiel.GameType.Utility.ZERO_SUM,
    reward_model=pyspiel.GameType.RewardModel.TERMINAL,
    max_num_players=_MAX_NUM_PLAYERS,
    min_num_players=_MIN_NUM_PLAYERS,
    provides_information_state_string=True,
    provides_information_state_tensor=True,
    provides_observation_string=False,
    provides_observation_tensor=True,
    parameter_specification={
        "players": _MIN_NUM_PLAYERS,
        "hand_length": _HAND_LENGTH,
        "num_digits": _NUM_DIGITS,
    },
)
_GAME_INFO = pyspiel.GameInfo(
    # Num actions = total number of cards * number of digits + action enum
    num_distinct_actions=_HAND_LENGTH * _NUM_DIGITS * _MIN_NUM_PLAYERS
    + BID_ACTION_OFFSET,
    max_chance_outcomes=_HAND_LENGTH * _NUM_DIGITS,
    num_players=_MIN_NUM_PLAYERS,
    min_utility=-(
        _MIN_NUM_PLAYERS - 1
    ),  # Reward from being challenged and losing.
    max_utility=_MIN_NUM_PLAYERS
    - 1,  # Reward for being challenged and winning.
    utility_sum=0.0,
    # Number of possible rounds: hand_length * num_digits * num_players
    # Total moves per round: num_players for non-rebid, num_players-1 for rebid
    # Max game length: number of possible rounds * total moves per round
    max_game_length=_HAND_LENGTH * _NUM_DIGITS * _MIN_NUM_PLAYERS**2,
)


class LiarsPoker(pyspiel.Game):
  """A Python version of Liar's poker."""

  def __init__(self, params=None):
    super().__init__(_GAME_TYPE, _GAME_INFO, params or dict())
    game_parameters = self.get_parameters()
    self.hand_length = game_parameters.get("hand_length", _HAND_LENGTH)
    self.num_digits = game_parameters.get("num_digits", _NUM_DIGITS)
    self.deck = _FULL_DECK[: self.num_digits]

  def new_initial_state(self):
    """Returns a state corresponding to the start of a game."""
    return LiarsPokerState(self)

  def make_py_observer(self, iig_obs_type=None, params=None):
    """Returns an object used for observing game state."""
    return LiarsPokerObserver(
        iig_obs_type or pyspiel.IIGObservationType(perfect_recall=False),
        self.num_players(),
        self.hand_length,
        self.num_digits,
        params,
    )


class LiarsPokerState(pyspiel.State):
  """A python version of the Liars Poker state."""

  def __init__(self, game):
    """Constructor; should only be called by Game.new_initial_state."""
    super().__init__(game)
    # Game attributes
    self._num_players = game.num_players()
    self._hand_length = game.hand_length
    self._num_digits = game.num_digits
    self._deck = game.deck
    self.hands = [[] for _ in range(self._num_players)]

    # Action dynamics
    self.total_possible_bids = (
        game.hand_length * game.num_digits * self._num_players
    )
    self.bid_history = np.zeros((self.total_possible_bids, self._num_players))
    self.challenge_history = np.zeros(
        (self.total_possible_bids, self._num_players)
    )
    # self._current_player is only the valid current_player when cards have
    # been dealt. Otherwise it's chance.
    self._current_player = 0
    self._max_bid = self._hand_length * self._num_digits * self._num_players
    self._bid_originator = -1
    self._current_action = -1
    self._num_challenges = 0
    self.is_rebid = False

    # Game over dynamics
    self._winner = -1
    self._loser = -1

  def current_player(self):
    """Returns id of the current player to act.

    The id is:
      - TERMINAL if game is over.
      - CHANCE if a player is drawing a number to fill out their hand.
      - a number otherwise.
    """
    if self.is_terminal():
      return pyspiel.PlayerId.TERMINAL
    elif len(self.hands[self._num_players - 1]) < self._hand_length:
      return pyspiel.PlayerId.CHANCE
    else:
      return self._current_player

  def winner(self):
    """Returns the id of the winner if the bid originator has won.

    -1 otherwise.
    """
    return self._winner

  def loser(self):
    """Returns the id of the loser if the bid originator has lost.

    -1 otherwise.
    """
    return self._loser

  def _is_challenge_possible(self):
    """A challenge is possible once the first bid is made."""
    return self._current_action != -1

  def _is_rebid_possible(self):
    """A rebid is only possible when all players have challenged the original bid."""
    return not self.is_rebid and self._num_challenges == self._num_players - 1

  def _legal_actions(self, player):
    """Returns a list of legal actions, sorted in ascending order."""
    assert player >= 0
    actions = []

    if self._is_challenge_possible():
      actions.append(CHALLENGE_ACTION)

    if player != self._bid_originator or self._is_rebid_possible():
      # Any move higher than the current bid is allowed.
      # Bids start at BID_ACTION_OFFSET (1) as 0 represents the challenge
      # action.
      for bid in range(
          max(BID_ACTION_OFFSET, self._current_action + 1), self._max_bid
      ):
        actions.append(bid)

    return actions

  def chance_outcomes(self):
    """Returns the possible chance outcomes and their probabilities."""
    assert self.is_chance_node()
    probability = 1.0 / self._num_digits
    return [(digit, probability) for digit in self._deck]

  def _decode_bid(self, bid):
    """Turns a bid ID to a (count, number) tuple.

    For example, take 2 players each with 2 numbers from the deck of 1, 2, and
    3.
      - A bid of two 1's would correspond to a bid id 1.
        - Explanation: 1 is the lowest number, and the only lower bid would be
        zero 1's.
      - A bid of three 3's would correspond to a bid id 10.
        - Explanation: 1-4 1's take bid ids 0-3. 1-4 2's take bid ids 4-7. 1 and
        2 3's take bid ids 8 and 9.

    Args:
      bid: Bid ID in the range 0 to self._max_bid (non-inclusive).

    Returns:
      A tuple of (count, number). For example, (1, 2) represents one 2's.
    """
    count = bid % (self._hand_length * self._num_players) + 1
    number = self._deck[bid // (self._hand_length * self._num_players)]
    return (count, number)

  def encode_bid(self, count, number):
    """Turns a count and number into a bid ID.

    Bid ID is in the range 0 to self._max_bid (non-inclusive).

    For example, take 2 players each with 2 numbers from the deck of 1, 2, and
    3.
      - A count of 2 and number of 1 would be a bid of two one's and a bid id 1.
        - Explanation: 1 is the lowest number, and the only lower bid would be
        zero 1's
          corresponding to bid id 0.

    Args:
      count: The count of the bid.
      number: The number of the bid.

    Returns:
      A single bid ID.
    """
    return ((number - 1) * self._hand_length * self._num_players) + count - 1

  def _counts(self):
    """Determines if the bid originator wins or loses."""
    bid_count, bid_number = self._decode_bid(
        self._current_action - BID_ACTION_OFFSET
    )

    # Count the number of bid_numbers from all players.
    matches = 0
    for player_id in range(self._num_players):
      for digit in self.hands[player_id]:
        if digit == bid_number:
          matches += 1

    # If the number of matches are at least the bid_count bid, then the bidder
    # wins. Otherwise everyone else wins.
    if matches >= bid_count:
      self._winner = self._bid_originator
    else:
      self._loser = self._bid_originator

  def _update_bid_history(self, bid, player):
    """Writes a player's bid into memory."""
    self.bid_history[bid][player] = 1

  def _update_challenge_history(self, bid, player):
    """Write a player's challenge for a bid into memory."""
    self.challenge_history[bid][player] = 1

  def _apply_action(self, action):
    """Applies an action and updates the state."""
    if self.is_chance_node():
      # If we are still populating hands, draw a number for the current player.
      self.hands[self._current_player].append(action)
    elif action == CHALLENGE_ACTION:
      assert self._is_challenge_possible()
      self._update_challenge_history(
          self._current_action - BID_ACTION_OFFSET, self._current_player
      )
      self._num_challenges += 1
      # If there is no ongoing rebid, check if all players challenge before
      # counting. If there is an ongoing rebid, count once all the players
      # except the bidder challenges.
      if (not self.is_rebid and self._num_challenges == self._num_players) or (
          self.is_rebid and self._num_challenges == self._num_players - 1
      ):
        self._counts()
    else:
      # Set the current bid to the action.
      self._current_action = action
      if self._current_player == self._bid_originator:
        # If the bid originator is bidding again, we have a rebid.
        self.is_rebid = True
      else:
        # Otherwise, we have a regular bid.
        self.is_rebid = False
      # Set the bid originator to the current player.
      self._bid_originator = self._current_player
      self._update_bid_history(
          self._current_action - BID_ACTION_OFFSET, self._current_player
      )
      self._num_challenges = 0
    self._current_player = (self._current_player + 1) % self._num_players

  def _action_to_string(self, player, action):
    """Action -> string."""
    if player == pyspiel.PlayerId.CHANCE:
      return f"Deal: {action}"
    elif action == CHALLENGE_ACTION:
      return "Challenge"
    else:
      count, number = self._decode_bid(action - BID_ACTION_OFFSET)
      return f"Bid: {count} of {number}"

  def is_terminal(self):
    """Returns True if the game is over."""
    return self._winner >= 0 or self._loser >= 0

  def returns(self):
    """Total reward for each player over the course of the game so far."""
    if self._winner != -1:
      bidder_reward = self._num_players - 1
      others_reward = -1.0
    elif self._loser != -1:
      bidder_reward = -1 * (self._num_players - 1)
      others_reward = 1.0
    else:
      # Game is not over.
      bidder_reward = 0.0
      others_reward = 0.0
    return [
        others_reward if player_id != self._bid_originator else bidder_reward
        for player_id in range(self._num_players)
    ]

  def __str__(self):
    """String for debug purposes. No particular semantics are required."""
    if self._current_action != -1:
      count, number = self._decode_bid(self._current_action - BID_ACTION_OFFSET)
    else:
      count, number = "None", "None"
    return (
        "Hands: {}, Bidder: {}, Current Player: {}, Current Bid: {} of {},"
        " Rebid: {}".format(
            self.hands,
            self._bid_originator,
            self.current_player(),
            count,
            number,
            self.is_rebid,
        )
    )


class LiarsPokerObserver:
  """Observer, conforming to the PyObserver interface (see observation.py).

  An observation will consist of the following:
    - One hot encoding of the current player number: [0 0 0 1 0 0 0]
    - A vector of length hand_length containing the digits in a player's hand.
    - Two matrices each of size (hand_length * num_digits * num_players,
    num_players)
      will store bids and challenges respectively. Each row in the matrix
      corresponds
      to a particular bid (e.g. one 1, two 5s, or eight 3s). 0 will represent no
      action. 1 will represent a player's bid or a player's challenge.
    - One bit for whether we are rebidding: [1] rebid occuring, [0] otherwise
    - One bit for whether we are counting: [1] COUNTS called, [0] otherwise
  """

  def __init__(
      self, iig_obs_type, num_players, hand_length, num_digits, params=None
  ):
    """Initiliazes an empty observation tensor."""
    del params
    self.num_players = num_players
    self.hand_length = hand_length

    # Determine which observation pieces we want to include.
    # Pieces is a list of tuples containing observation pieces.
    # Pieces are described by their (name, number of elements, and shape).
    pieces = [(
        "player",
        num_players,
        (num_players,),
    )]  # One-hot encoding for the player id.
    if iig_obs_type.private_info == pyspiel.PrivateInfoType.SINGLE_PLAYER:
      # Vector containing the digits in a player's hand
      pieces.append(("private_hand", hand_length, (hand_length,)))
    if iig_obs_type.public_info:
      pieces.append(("rebid_state", 1, (1,)))
      pieces.append(("counts_state", 1, (1,)))
      if iig_obs_type.perfect_recall:
        # One-hot encodings for players' moves at every round.
        total_possible_rounds = hand_length * num_digits * num_players
        pieces.append((
            "bid_history",
            total_possible_rounds * num_players,
            (total_possible_rounds, num_players),
        ))
        pieces.append((
            "challenge_history",
            total_possible_rounds * num_players,
            (total_possible_rounds, num_players),
        ))

    # Build the single flat tensor.
    total_size = sum(size for name, size, shape in pieces)
    self.tensor = np.zeros(total_size, np.float32)

    # Build the named & reshaped views of the bits of the flat tensor.
    self.dict = {}
    index = 0
    for name, size, shape in pieces:
      self.dict[name] = self.tensor[index : index + size].reshape(shape)
      index += size

  def set_from(self, state, player):
    """Updates `tensor` and `dict` to reflect `state` from PoV of `player`."""
    self.tensor.fill(0)
    if "player" in self.dict:
      self.dict["player"][player] = 1
    if (
        "private_hand" in self.dict
        and len(state.hands[player]) == self.hand_length
    ):
      self.dict["private_hand"] = np.asarray(state.hands[player])
    if "rebid_state" in self.dict:
      self.dict["rebid_state"][0] = int(state.is_rebid)
    if "counts_state" in self.dict:
      self.dict["counts_state"][0] = int(state.is_terminal())
    if "bid_history" in self.dict:
      self.dict["bid_history"] = state.bid_history
    if "challenge_history" in self.dict:
      self.dict["challenge_history"] = state.challenge_history

  def string_from(self, state, player):
    """Observation of `state` from the PoV of `player`, as a string."""
    pieces = []
    if "player" in self.dict:
      pieces.append(f"p{player}")
    if (
        "private_hand" in self.dict
        and len(state.hands[player]) == self.hand_length
    ):
      pieces.append(f"hand:{state.hands[player]}")
    if "rebid_state" in self.dict:
      pieces.append(f"rebid:{[int(state.is_rebid)]}")
    if "counts_state" in self.dict:
      pieces.append(f"counts:{[int(state.is_terminal())]}")
    if "bid_history" in self.dict:
      for bid in range(len(state.bid_history)):
        if np.any(state.bid_history[bid] == 1):
          pieces.append("b:{}.".format(bid))
    if "challenge_history" in self.dict:
      for bid in range(len(state.challenge_history)):
        if np.any(state.challenge_history[bid] == 1):
          pieces.append("c:{}.".format(bid))
    return " ".join(str(p) for p in pieces)


# Register the game with the OpenSpiel library

pyspiel.register_game(_GAME_TYPE, LiarsPoker)
